"""Traffic APIs."""

from .intermediate_traffic import IntermediateTrafficApi
from .junction_analytics import JunctionAnalyticsApi
from .od_analytics import ODAnalysisApi
from .route_monitoring import RouteMonitoringApi
from .traffic import TrafficApi
from .traffic_stats import TrafficStatsApi

__all__ = [
    "IntermediateTrafficApi",
    "JunctionAnalyticsApi",
    "ODAnalysisApi",
    "RouteMonitoringApi",
    "TrafficStatsApi",
    "TrafficApi",
]



"""Waypoint optimization API."""

from ..api import BaseApi, BaseParams
from ..routing.models import WaypointOptimizationPostData, WaypointOptimizedResponse


class WaypointOptimizationApi(BaseApi):
    """Waypoint optimization API.

    TomTom's Waypoint Optimization service is intended to optimize the order of provided waypoints by fastest route. This service uses an heuristic
    algorithm to create an optimized sequence.

    For more information, see: https://developer.tomtom.com/waypoint-optimization/documentation/waypoint-optimization-service
    """

    async def post_waypointoptimization(
        self,
        *,
        params: BaseParams | None = None,  # No extra params.
        data: WaypointOptimizationPostData,
    ) -> WaypointOptimizedResponse:
        """Post waypointoptimization.

        For more information, see: https://developer.tomtom.com/waypoint-optimization/documentation/waypoint-optimization

        Args:
            params (BaseParams | None, optional): Optional parameters for the request. Defaults to None.
            data (WaypointOptimizationPostData): Data specifying the waypoints and any optimization criteria.

        Returns:
            WaypointOptimizedResponse: The response containing the optimized order of waypoints.
        """
        response = await self.post(
            endpoint="/routing/waypointoptimization/1",
            params=params,
            data=data,
        )

        return await response.deserialize(WaypointOptimizedResponse)



"""Geocode API."""

from ..api import BaseApi
from ..places.models import GeocodeParams, SearchResponse, StructuredGeocodeParams


class GeocodingApi(BaseApi):
    """Geocoding API.

    The Geocoding API is a powerful tool that converts addresses, such as "109 Park Row, New York, United States," into geographic coordinates (e.g.,
    "lat": 40.71226, "lon": -74.00207). Designed for machine-to-machine interaction, the TomTom Geocoding API is capable of handling requests from
    automated systems to geocode addresses that may be incomplete, incorrectly formatted, or contain typos, providing the best possible result.

    For more information, see: https://developer.tomtom.com/geocoding-api/documentation/product-information/introduction
    """

    async def get_geocode(
        self,
        *,
        query: str,
        params: GeocodeParams | None = None,
    ) -> SearchResponse:
        """Get geocode.

        For more information, see: https://developer.tomtom.com/geocoding-api/documentation/geocode

        Args:
            query (str): The query string representing the address or place to geocode.
            params (GeocodeParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            SearchResponse: Response containing search results.
        """
        response = await self.get(
            endpoint=f"/search/2/geocode/{query}.json",
            params=params,
        )

        return await response.deserialize(SearchResponse)

    async def get_structured_geocode(
        self,
        *,
        countryCode: str,
        params: StructuredGeocodeParams | None = None,
    ) -> SearchResponse:
        """Get structured geocode.

        For more information, see: https://developer.tomtom.com/geocoding-api/documentation/structured-geocode

        Args:
            countryCode (str): The country code representing the location's country (e.g., "US" for the United States).
            params (StructuredGeocodeParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            SearchResponse: Response containing search results.
        """
        response = await self.get(
            endpoint=f"/search/2/structuredGeocode.json?countryCode={countryCode}",
            params=params,
        )

        return await response.deserialize(SearchResponse)



"""EV Search API."""

from ..api import BaseApi
from .models import EvSearchByIdParams, EvSearchNearbyParams, SearchResponse


class EVSearchApi(BaseApi):
    """EV Search API.

    EV Search is a REST API limited to and optimized for electric vehicle station POI category. It provides complete EV POI data including static
    location information (lat/long), address, opening hours, access restrictions, technical specs of the charging station (connector type, voltage,
    power, current, current type), etc. as well as dynamic availability status. Wide range of EV specific filters (including dynamic availability)
    allows for narrowing the search results to match the personal preference (like charging power or access type) of the driver or the technical
    specification of the electric vehicle (connector type).

    For more information, see: https://developer.tomtom.com/ev-search-api/documentation/product-information/introduction
    """

    async def get_ev_search_nearby(
        self,
        *,
        params: EvSearchNearbyParams | None = None,
    ) -> SearchResponse:
        """Get EV search nearby.

        For more information, see: https://developer.tomtom.com/ev-search-api/documentation/ev-search-api/ev-search-nearby

        Args:
            params (EvSearchNearbyParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            SearchResponse: Response containing search results.
        """
        response = await self.get(
            endpoint="/search/2/evsearch",
            params=params,
        )

        return await response.deserialize(SearchResponse)

    async def get_ev_search_by_id(
        self,
        *,
        params: EvSearchByIdParams | None = None,
    ) -> SearchResponse:
        """Get EV search by id.

        For more information, see: https://developer.tomtom.com/ev-search-api/documentation/ev-search-api/ev-search-by-id

        Args:
            params (EvSearchByIdParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            SearchResponse: Response containing search results.
        """
        response = await self.get(
            endpoint="/search/2/evbyid",
            params=params,
        )

        return await response.deserialize(SearchResponse)



"""Tracking & Logistics APIs."""

from .geofencing import GeofencingApi
from .location_history import LocationHistoryApi
from .notifications import NotificationsApi
from .snap_to_roads import SnapToRoadsApi

__all__ = [
    "GeofencingApi",
    "LocationHistoryApi",
    "NotificationsApi",
    "SnapToRoadsApi",
]



"""Models for the TomTom Places API."""
# pylint: disable=invalid-name, too-many-instance-attributes, too-many-lines

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date
from enum import StrEnum

from geojson import Feature, FeatureCollection, GeometryCollection, LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon
from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from ..api import BaseParams, BasePostData
from ..models import Language, LatLon, ViewType


class AccessType(StrEnum):
    """Supported access types."""

    PUBLIC = "Public"
    AUTHORIZED = "Authorized"
    RESTRICTED = "Restricted"
    PRIVATE = "Private"


@dataclass(kw_only=True)
class AdditionalDataItem:
    """Represents an additional data response item."""

    providerID: str
    error: str | None = None
    geometryData: (
        Feature | FeatureCollection | GeometryCollection | LineString | MultiLineString | MultiPoint | MultiPolygon | Point | Polygon | None
    ) = None


@dataclass(kw_only=True)
class AdditionalDataParams(BaseParams):
    """Parameters for the get_additional_data method."""

    geometriesZoom: int | None = None


@dataclass(kw_only=True)
class AdditionalDataResponse(DataClassORJSONMixin):
    """Represents a Additional Data response."""

    additionalData: list[AdditionalDataItem]


@dataclass(kw_only=True)
class Address(DataClassORJSONMixin):
    """Represents a Address."""

    streetNumber: str | None = None
    streetName: str | None = None
    municipalitySubdivision: str | None = None
    municipalitySecondarySubdivision: str | None = None
    neighbourhood: str | None = None
    municipality: str | None = None
    countrySecondarySubdivision: str | None = None
    countryTertiarySubdivision: str | None = None
    countrySubdivision: str | None = None
    postalCode: str | None = None
    postalName: str | None = None
    extendedPostalCode: str | None = None
    countryCode: str | None = None
    country: str | None = None
    countryCodeISO3: str | None = None
    freeformAddress: str | None = None
    countrySubdivisionName: str | None = None
    countrySubdivisionCode: str | None = None
    localName: str | None = None


@dataclass(kw_only=True)
class AddressRange(DataClassORJSONMixin):
    """Represents a AddressRange."""

    rangeLeft: str
    rangeRight: str
    from_: LatLon = field(metadata=field_options(alias="from"))
    to: LatLon


@dataclass(kw_only=True)
class Addresses(DataClassORJSONMixin):
    """Represents Addresses."""

    address: RevGeocodeAddress
    position: str
    roadClass: list[Roadclass] | None = None
    # roadUse:  # Deprecated
    matchType: MatchType | None = None
    dataSources: DataSources | None = None
    entityType: list[EntityType] | None = None
    mapcodes: list[MapCode] | None = None
    id: str | None = None


@dataclass(kw_only=True)
class AsynchronousBatchDownloadParams(BaseParams):
    """Parameters for the get_asynchronous_batch_download method."""

    waitTimeSeconds: int


@dataclass(kw_only=True)
class AsynchronousSynchronousBatchParams(BaseParams):
    """Parameters for the post_asynchronous_batch_submission method."""

    redirectMode: RedirectModeType
    waitTimeSeconds: int


@dataclass(kw_only=True)
class AutocompleteParams(BaseParams):
    """Parameters for the get_autocomplete method."""

    limit: int | None = None
    lat: float | None = None
    lon: float | None = None
    radius: int | None = None
    countrySet: list[str] | None = None
    resultSet: list[ResultSetType] | None = None


@dataclass(kw_only=True)
class AutocompleteResponse(DataClassORJSONMixin):
    """Represents an auto complete response."""

    context: Context
    results: list[AutocompleteResult]


@dataclass(kw_only=True)
class AutocompleteResult(DataClassORJSONMixin):
    """Represents an auto complete result."""

    segments: list[Segment]


@dataclass(kw_only=True)
class BatchItem:
    """Represents an BatchItem."""

    query: str
    post: list[Geometry | Route] | None = None


@dataclass(kw_only=True)
class BatchItemResponse(DataClassORJSONMixin):
    """Represents a Batch Item response."""

    statusCode: int
    response: Response


@dataclass(kw_only=True)
class BatchPostData(BasePostData):
    """Data for the post Batch API."""

    batchItems: list[BatchItem]


@dataclass(kw_only=True)
class BatchResponse(DataClassORJSONMixin):
    """Represents a Batch response."""

    formatVersion: str
    batchItems: list[BatchItemResponse]
    summary: BatchResponseSummary


@dataclass(kw_only=True)
class BatchResponseSummary(DataClassORJSONMixin):
    """Represents a batch response summary."""

    successfulRequests: int
    totalRequests: int


@dataclass(kw_only=True)
class BookmarkQueryIntent(DataClassORJSONMixin):
    """Represents a BookmarkQueryIntent."""

    bookmark: BookmarkType


class BookmarkType(StrEnum):
    """Supported bookmark types."""

    HOME = "HOME"
    WORK = "WORK"


@dataclass(kw_only=True)
class BoundingBox(DataClassORJSONMixin):
    """Represents a BoundingBox."""

    topLeftPoint: LatLon
    btmRightPoint: LatLon


@dataclass(kw_only=True)
class Brand(DataClassORJSONMixin):
    """Represents a Brand."""

    name: str


class CapabilitieType(StrEnum):
    """Supported capabilities."""

    CHARGING_PROFILE_CAPABLE = "ChargingProfileCapable"
    CHARGING_PREFERENCES_CAPABLE = "ChargingPreferencesCapable"
    CHIP_CARD_SUPPORT = "ChipCardSupport"
    CONTACTLESS_CARD_SUPPORT = "ContactlessCardSupport"
    CREDIT_CARD_PAYABLE = "CreditCardPayable"
    DEBIT_CARD_PAYABLE = "DebitCardPayable"
    PED_TERMINAL = "PedTerminal"
    REMOTE_START_STOP_CAPABLE = "RemoteStartStopCapable"
    RESERVABLE = "Reservable"
    RFID_READER = "RfidReader"
    START_SESSION_CONNECTOR_REQUIRED = "StartSessionConnectorRequired"
    TOKEN_GROUP_CAPABLE = "TokenGroupCapable"
    UNLOCK_CAPABLE = "UnlockCapable"
    PLUG_AND_CHARGE = "PlugAndCharge"


@dataclass(kw_only=True)
class CategorySearchParams(BaseParams):
    """Parameters for the get_category_search method."""

    typeahead: bool | None = None
    limit: int | None = None
    ofs: int | None = None
    countrySet: list[str] | None = None
    lat: float | None = None
    lon: float | None = None
    radius: int | None = None
    topLeft: str | None = None
    btmRight: str | None = None
    geoBias: str | None = None
    language: Language | None = None
    extendedPostalCodesFor: list[ExtendedPostalCodesForType] | None = None
    categorySet: list[str] | None = None
    brandSet: list[str] | None = None
    connectorSet: list[ConnectorType] | None = None
    minPowerKW: float | None = None
    maxPowerKW: float | None = None
    fuelSet: list[FuelType] | None = None
    vehicleTypeSet: list[VehicleType] | None = None
    view: ViewType | None = None
    openingHours: OpeningHoursType | None = None
    mapcodes: list[MapCodeType] | None = None
    timeZone: str | None = None
    relatedPois: RelatedPoisType | None = None


@dataclass(kw_only=True)
class ChargingPark(DataClassORJSONMixin):
    """Represents a ChargingPark."""

    connectors: list[Connector]


@dataclass(kw_only=True)
class ChargingPoint(DataClassORJSONMixin):
    """Represents a ChargingPoint."""

    capabilities: list[CapabilitieType]
    connectors: list[Connector]
    evseId: str
    physicalReference: str | None = None
    status: StatusType


@dataclass(kw_only=True)
class ChargingStation(DataClassORJSONMixin):
    """Represents a ChargingStation."""

    id: str
    chargingPoints: list[ChargingPoint]


@dataclass(kw_only=True)
class Classification(DataClassORJSONMixin):
    """Represents a Classification."""

    code: str
    names: list[Name]


@dataclass(kw_only=True)
class Connector(DataClassORJSONMixin):
    """Represents a Connector.

    Note that some API's use the field type while others use connectorType, after deserialization the values for both fields will be the same.
    """

    id: str | None = None
    type: ConnectorType | None = None  # In EV search.
    connectorType: ConnectorType | None = None  # In general search.
    ratedPowerKW: float
    voltageV: int | None = None
    currentA: int | None = None
    currentType: CurrentType

    @classmethod
    def __post_deserialize__(cls, obj: Connector) -> Connector:
        """Connector post deserialize.

        Handle the discrepancy between general search and EV search on connector type.
        """
        if obj.type is not None and obj.connectorType is None:
            obj.connectorType = obj.type
        elif obj.connectorType is not None and obj.type is None:
            obj.type = obj.connectorType
        return obj


class ConnectorType(StrEnum):
    """Supported connector types."""

    STANDARD_HOUSEHOLD_COUNTRY_SPECIFIC = "StandardHouseholdCountrySpecific"
    IEC62196_TYPE1 = "IEC62196Type1"
    IEC62196_TYPE1_CCS = "IEC62196Type1CCS"
    IEC62196_TYPE2_CABLE_ATTACHED = "IEC62196Type2CableAttached"
    IEC62196_TYPE2_OUTLET = "IEC62196Type2Outlet"
    IEC62196_TYPE2_CCS = "IEC62196Type2CCS"
    IEC62196_TYPE3 = "IEC62196Type3"
    CHADEMO = "Chademo"
    GBT20234_PART2 = "GBT20234Part2"
    GBT20234_PART3 = "GBT20234Part3"
    IEC60309_AC3_PHASE_RED = "IEC60309AC3PhaseRed"
    IEC60309_AC1_PHASE_BLUE = "IEC60309AC1PhaseBlue"
    IEC60309_DC_WHITE = "IEC60309DCWhite"
    TESLA = "Tesla"


@dataclass(kw_only=True)
class Context(DataClassORJSONMixin):
    """Represents an auto complete context."""

    inputQuery: str
    geoBias: GeoBias | None = None


@dataclass(kw_only=True)
class CoordinateQueryIntent(DataClassORJSONMixin):
    """Represents a CoordinateQueryIntent."""

    lat: float
    lon: float


@dataclass(kw_only=True)
class CrossStreetLookupParams(BaseParams):
    """Parameters for the get_cross_street_lookup method."""

    limit: int | None = None
    radius: int | None = None
    language: Language | None = None
    allowFreeformNewLine: bool | None = None
    view: ViewType | None = None


class CurrentType(StrEnum):
    """Supported current types."""

    AC1 = "AC1"
    AC3 = "AC3"
    DC = "DC"


@dataclass(kw_only=True)
class DataSources(DataClassORJSONMixin):
    """Represents a DataSources."""

    chargingAvailability: IdString


class EntityType(StrEnum):
    """Supported entity types."""

    COUNTRY = "Country"
    COUNTRY_SUBDIVISION = "CountrySubdivision"
    COUNTRY_SECONDARY_SUBDIVISION = "CountrySecondarySubdivision"
    COUNTRY_TERTIARY_SUBDIVISION = "CountryTertiarySubdivision"
    MUNICIPALITY = "Municipality"
    MUNICIPALITY_SUBDIVISION = "MunicipalitySubdivision"
    MUNICIPALITY_SECONDARY_SUBDIVISION = "MunicipalitySecondarySubdivision"
    NEIGHBOURHOOD = "Neighbourhood"
    POSTAL_CODE_AREA = "PostalCodeArea"


@dataclass(kw_only=True)
class EntryPoint(DataClassORJSONMixin):
    """Represents a EntryPoint."""

    type: EntryPointType
    functions: list[FunctionType] | None = None
    pathToNext: PathToNextType | None = None
    position: LatLon


class EntryPointType(StrEnum):
    """Supported entry point types."""

    MAIN = "main"
    MINOR = "minor"
    ROUTE = "route"


@dataclass(kw_only=True)
class EvSearchByIdParams(BaseParams):
    """Parameters for the get_ev_search_by_id method."""

    id: str


@dataclass(kw_only=True)
class EvSearchNearbyParams(BaseParams):
    """Parameters for the get_ev_search_nearby method."""

    lat: float
    lon: float
    radius: int
    limit: int | None = None
    status: list[StatusType] | None = None
    connector: list[ConnectorType] | None = None
    accessType: list[AccessType] | None = None
    restriction: list[RestrictionType] | None = None
    capability: list[CapabilitieType] | None = None
    minPowerKW: float | None = None
    maxPowerKW: float | None = None


class ExtendedPostalCodesForType(StrEnum):
    """Supported extended postal codes."""

    ADDR = "Addr"
    PAD = "PAD"
    POI = "POI"
    NONE = "None"


class FilterType(StrEnum):
    """Supported filter types."""

    BACK_ROADS = "BackRoads"


class FuelType(StrEnum):
    """Supported fuel types.

    For more information, see: https://developer.tomtom.com/search-api/documentation/product-information/supported-fuel-types.
    """

    PETROL = "Petrol"
    LPG = "LPG"
    DIESEL = "Diesel"
    BIO_DIESEL = "Biodiesel"
    DIESEL_FOR_COMMERCIAL_VEHICLES = "DieselForCommercialVehicles"
    E85 = "E85"
    LNG = "LNG"
    CNG = "CNG"
    HYDROGEN = "Hydrogen"
    AD_BLUE = "AdBlue"


class FunctionType(StrEnum):
    """Represents the type of access for the Address."""

    MAIN = "Main"
    POSTAL = "Postal"
    ROUTING = "Routing"
    EMERGENCY = "Emergency"
    PEDESTRIAN = "Pedestrian"
    DELIVERY = "Delivery"
    AUTHORIZED = "Authorized"
    FRONT_DOOR = "FrontDoor"
    # Premium content:
    PARKING = "PARKING"
    ENTRANCE = "ENTRANCE"
    ELEVATOR = "ELEVATOR"
    STAIR = "STAIR"


@dataclass(kw_only=True)
class GeoBias(DataClassORJSONMixin):
    """Represents a geo bias."""

    position: LatLon
    radius: int


@dataclass(kw_only=True)
class GeocodeParams(BaseParams):
    """Represents the parameters for get_geocode."""

    storeResult: bool | None = None  # Deprecated
    typeahead: bool | None = None  # Deprecated
    limit: int | None = None
    ofs: int | None = None
    lat: float | None = None
    lon: float | None = None
    countrySet: str | None = None
    radius: int | None = None
    topLeft: str | None = None
    btmRight: str | None = None
    language: Language | None = None
    extendedPostalCodesFor: list[ExtendedPostalCodesForType] | None = None
    view: ViewType | None = None
    mapcodes: list[MapCodeType] | None = None
    entityTypeSet: list[EntityType] | None = None


@dataclass(kw_only=True)
class Geometry:
    """Represents an geometry."""

    type: str
    position: str | None = None
    radius: int | None = None
    vertices: list[str] | None = None


@dataclass(kw_only=True)
class GeometryFilterData(BasePostData):
    """Data for the post geometry filter API."""

    geometryList: list[Geometry]
    poiList: list[GeometryPoi]


@dataclass(kw_only=True)
class GeometryFilterResponse(DataClassORJSONMixin):
    """Represents a Geometry Filter response."""

    summary: Summary
    results: list[GeometryPoi]


@dataclass(kw_only=True)
class GeometryPoi:
    """Represents an geometry poi."""

    poi: Poi | None = None
    address: Address | None = None
    position: LatLon


@dataclass(kw_only=True)
class GeometrySearchParams(BaseParams):
    """Parameters for the get_geometry_search method."""

    limit: int | None = None
    language: Language | None = None
    lat: float | None = None
    lon: float | None = None
    extendedPostalCodesFor: list[ExtendedPostalCodesForType] | None = None
    idxSet: list[IdxSetType] | None = None
    categorySet: list[str] | None = None
    brandSet: list[str] | None = None
    connectorSet: list[ConnectorType] | None = None
    minPowerKW: float | None = None
    maxPowerKW: float | None = None
    fuelSet: list[FuelType] | None = None
    vehicleTypeSet: list[VehicleType] | None = None
    view: ViewType | None = None
    openingHours: OpeningHoursType | None = None
    timeZone: str | None = None
    mapcodes: list[MapCodeType] | None = None
    relatedPois: RelatedPoisType | None = None
    entityTypeSet: list[EntityType] | None = None


@dataclass(kw_only=True)
class GeometrySearchPostData(BasePostData):
    """Data for the post geometry search API."""

    geometryList: list[Geometry]


@dataclass(kw_only=True)
class Id(DataClassORJSONMixin):
    """Represents an Id."""

    id: int


@dataclass(kw_only=True)
class IdString(DataClassORJSONMixin):
    """Represents an IdString."""

    id: str


class IdxSetType(StrEnum):
    """Supported IdxSet types.

    For more information, see: https://developer.tomtom.com/search-api/documentation/search-service/fuzzy-search#indexes-abbreviation-values.
    """

    GEO = "Geo"
    PAD = "PAD"
    ADDR = "Addr"
    STR = "Str"
    XSTR = "XStr"
    POI = "POI"


@dataclass(kw_only=True)
class MapCode(DataClassORJSONMixin):
    """Represents a MapCode."""

    type: MapCodeType
    fullMapcode: str
    territory: str
    code: str


class MapCodeType(StrEnum):
    """A mapcode represents a specific location to within a few meters.

    Every location on Earth can be represented by a mapcode. Mapcodes are designed to be short, easy to recognize, remember, and communicate.
    Visit the Mapcode project website for more information.
    See: http://www.mapcode.com/.
    """

    LOCAL = "Local"
    INTERNATIONAL = "International"
    ALTERNATIVE = "Alternative"


@dataclass(kw_only=True)
class Match(DataClassORJSONMixin):
    """Represents a match."""

    offset: int
    length: int


class MatchType(StrEnum):
    """Supported match types."""

    ADDRESS_POINT = "AddressPoint"
    HOUSE_NUMBER_RANGE = "HouseNumberRange"
    STREET = "Street"


@dataclass(kw_only=True)
class Matches(DataClassORJSONMixin):
    """Represents a matches."""

    inputQuery: list[Match]


@dataclass(kw_only=True)
class Name(DataClassORJSONMixin):
    """Represents a Name."""

    nameLocale: str
    name: str


@dataclass(kw_only=True)
class NearbyQueryIntent(DataClassORJSONMixin):
    """Represents a NearbyQueryIntent."""

    lat: float
    lon: float
    query: str
    text: str


@dataclass(kw_only=True)
class NearbySearchParams(BaseParams):
    """Parameters for the get_nearby_search method."""

    limit: int | None = None
    ofs: int | None = None
    countrySet: list[str] | None = None
    radius: int | None = None
    topLeft: str | None = None
    btmRight: str | None = None
    language: Language | None = None
    extendedPostalCodesFor: list[ExtendedPostalCodesForType] | None = None
    categorySet: list[str] | None = None
    brandSet: list[str] | None = None
    connectorSet: list[ConnectorType] | None = None
    minPowerKW: float | None = None
    maxPowerKW: float | None = None
    fuelSet: list[FuelType] | None = None
    vehicleTypeSet: list[VehicleType] | None = None
    view: ViewType | None = None
    openingHours: OpeningHoursType | None = None
    mapcodes: list[MapCodeType] | None = None
    timeZone: str | None = None
    relatedPois: RelatedPoisType | None = None


@dataclass(kw_only=True)
class OpeningHour(DataClassORJSONMixin):
    """Represents an OpeningHour."""

    mode: OpeningHoursType
    timeRanges: list[TimeRange]


class OpeningHoursType(StrEnum):
    """Supported opening hours type."""

    NEXT_SEVEN_DAYS = "nextSevenDays"


class PathToNextType(StrEnum):
    """Supported path to next types."""

    DRIVING = "DRIVING"
    WALKING = "WALKING"
    UNKNOWN = "UNKNOWN"


@dataclass(kw_only=True)
class PaymentOption(DataClassORJSONMixin):
    """Represents a PaymentOption."""

    brands: list[Brand]


@dataclass(kw_only=True)
class PlaceByIdParams(BaseParams):
    """Parameters for the get_place_by_id method."""

    entityId: str
    language: Language | None = None
    openingHours: OpeningHoursType | None = None
    timeZone: str | None = None
    mapcodes: list[MapCodeType] | None = None
    relatedPois: RelatedPoisType | None = None
    view: ViewType | None = None


@dataclass(kw_only=True)
class PlaceByIdResponse(DataClassORJSONMixin):
    """Represents a Place by Id response."""

    summary: Summary
    results: list[Result]


@dataclass(kw_only=True)
class Poi(DataClassORJSONMixin):
    """Represents a Result."""

    name: str
    phone: str | None = None
    brands: list[Brand] | None = None
    url: str | None = None
    # categories: list[str]  # Deprecated: Use classifications instead. As of May 1, 2018
    categorySet: list[Id] | None = None
    openingHours: list[OpeningHour] | None = None
    classifications: list[Classification] | None = None
    timeZone: TimeZone | None = None


@dataclass(kw_only=True)
class PoiCategoriesParams(BaseParams):
    """Represents the parameters for get_poi_categories."""

    language: Language | None = None


@dataclass(kw_only=True)
class PoiCategoriesResponse(DataClassORJSONMixin):
    """Represents a Poi Categories response."""

    poiCategories: list[PoiCategory] | None = None


@dataclass(kw_only=True)
class PoiCategory(DataClassORJSONMixin):
    """Represents a PoiCategory."""

    id: int
    name: str
    childCategoryIds: list[int]
    synonyms: list[str]


@dataclass(kw_only=True)
class PoiSearchParams(BaseParams):
    """Parameters for the get_poi_search method."""

    typeahead: bool | None = None
    limit: int | None = None
    ofs: int | None = None
    countrySet: list[str] | None = None
    lat: float | None = None
    lon: float | None = None
    radius: int | None = None
    topLeft: str | None = None
    btmRight: str | None = None
    geoBias: str | None = None
    language: Language | None = None
    extendedPostalCodesFor: list[ExtendedPostalCodesForType] | None = None
    categorySet: list[str] | None = None
    brandSet: list[str] | None = None
    connectorSet: list[ConnectorType] | None = None
    minPowerKW: float | None = None
    maxPowerKW: float | None = None
    fuelSet: list[FuelType] | None = None
    vehicleTypeSet: list[VehicleType] | None = None
    view: ViewType | None = None
    openingHours: OpeningHoursType | None = None
    mapcodes: list[MapCodeType] | None = None
    timeZone: str | None = None
    relatedPois: RelatedPoisType | None = None


@dataclass(kw_only=True)
class Points:
    """Represents route points."""

    points: list[LatLon]


@dataclass(kw_only=True)
class PremiumGeocodeParams(BaseParams):
    """Parameters for the premium get_geocode method."""

    unit: str | None = None
    limit: int | None = None
    ofs: int | None = None
    lat: float | None = None
    lon: float | None = None
    countrySet: list[str] | None = None
    radius: int | None = None
    topLeft: str | None = None
    btmRight: str | None = None
    language: Language | None = None
    extendedPostalCodesFor: list[ExtendedPostalCodesForType] | None = None
    view: ViewType | None = None
    mapcodes: list[MapCodeType] | None = None
    entityTypeSet: list[EntityType] | None = None


@dataclass(kw_only=True)
class QueryIntent(DataClassORJSONMixin):
    """Represents a QueryIntent."""

    type: QueryIntentType
    details: CoordinateQueryIntent | NearbyQueryIntent | W3WQueryIntent | BookmarkQueryIntent


class QueryIntentType(StrEnum):
    """Supported query intent types."""

    COORDINATE = "COORDINATE"
    NEARBY = "NEARBY"
    W3W = "W3W"
    BOOKMARK = "BOOKMARK"


class QueryType(StrEnum):
    """Supported query types."""

    NEARBY = "NEARBY"
    NON_NEAR = "NON_NEAR"


class RedirectModeType(StrEnum):
    """Supported redirect mode types."""

    AUTO = "auto"
    MANUAL = "manual"


@dataclass(kw_only=True)
class RelatedPoi(DataClassORJSONMixin):
    """Represents a RelatedPoi."""

    relationType: RelationType
    id: str


class RelatedPoisType(StrEnum):
    """Supported related pois type."""

    ALL = "all"
    CHILD = "child"
    OFF = "off"
    PARENT = "parent"


class RelationType(StrEnum):
    """Supported relation types."""

    CHILD = "child"
    PARENT = "parent"


@dataclass(kw_only=True)
class Response(DataClassORJSONMixin):
    """Represents a Batch Item Response."""

    summary: Summary | None = None
    results: list[Result] | None = None
    errorText: str | None = None
    message: str | None = None
    httpStatusCode: int | None = None


class RestrictionType(StrEnum):
    """Supported parking restrictions."""

    EV_ONLY = "evOnly"
    PLUGGED = "plugged"
    DISABLED = "disabled"
    CUSTOMERS = "customers"
    MOTORCYCLES = "motorcycles"


@dataclass(kw_only=True)
class Result(DataClassORJSONMixin):
    """Represents a Result."""

    type: ResultType | None = None
    id: str
    score: float | None = None
    dist: float | None = None
    info: str | None = None
    entityType: EntityType | None = None
    poi: Poi | None = None
    relatedPois: list[RelatedPoi] | None = None
    address: Address
    position: LatLon
    mapcodes: list[MapCode] | None = None
    viewport: Viewport | None = None
    boundingBox: BoundingBox | None = None
    entryPoints: list[EntryPoint] | None = None
    detourTime: int | None = None
    detourDistance: int | None = None
    detourOffset: int | None = None
    addressRanges: list[AddressRange] | None = None
    chargingPark: ChargingPark | None = None
    dataSources: DataSources | None = None
    fuelTypes: list[FuelType] | None = None
    vehicleTypes: list[VehicleType] | None = None
    chargingStations: list[ChargingStation] | None = None
    openingHours: OpeningHour | None = None
    timeZone: TimeZone | None = None
    paymentOptions: list[PaymentOption] | None = None
    accessType: AccessType | None = None


class ResultSetType(StrEnum):
    """Supported result set types."""

    CATEGORY = "category"
    BRAND = "brand"


class ResultType(StrEnum):
    """Supported result types."""

    POI = "POI"
    STREET = "Street"
    GEOGRAPHY = "Geography"
    POINT_ADDRESS = "Point Address"
    ADDRESS = "Address"
    ADDRESS_RANGE = "Address Range"
    CROSS_STREET = "Cross Street"


@dataclass(kw_only=True)
class RevGeoBoundingBox(DataClassORJSONMixin):
    """Represents a BoundingBox for reverse geocode."""

    northEast: str
    southWest: str
    entity: RevGeoEntityType


class RevGeoEntityType(StrEnum):
    """Supported rev geo entity types."""

    POSITION = "position"


@dataclass(kw_only=True)
class RevGeocodeAddress(DataClassORJSONMixin):
    """Represents a Reverse Geocode Address."""

    buildingNumber: str | None = None  # Deprecated
    building: str | None = None
    streetNumber: str | None = None
    routeNumbers: list | None = None
    street: str | None = None  # Deprecated
    streetName: str
    crossStreet: str | None = None
    streetNameAndNumber: str | None = None
    speedLimit: str | None = None
    countryCode: str
    countrySubdivision: str | None = None
    countrySecondarySubdivision: str | None = None
    countryTertiarySubdivision: str | None = None
    municipality: str
    postalName: str | None = None
    postalCode: str | None = None
    municipalitySubdivision: str | None = None
    municipalitySecondarySubdivision: str | None = None
    neighbourhood: str | None = None
    sideOfStreet: str | None = None  # Deprecated
    offsetPosition: str | None = None  # Deprecated
    country: str
    countryCodeISO3: str
    freeformAddress: str
    boundingBox: RevGeoBoundingBox | None = None
    countrySubdivisionName: str | None = None
    countrySubdivisionCode: str | None = None
    localName: str | None = None


@dataclass(kw_only=True)
class ReverseGeocodeParams(BaseParams):
    """Parameters for the get_reverse_geocode method."""

    returnSpeedLimit: bool | None = None
    heading: float | None = None
    radius: int | None = None
    # number Deprecated: Support for the number parameter will be removed.
    # returnRoadClass Deprecated: Right now, the service supports only a single value, but in future, it may change.
    # returnRoadUse Deprecated: Support for the returnRoadUse parameter will be removed.
    # roadUse Deprecated: Support for the roadUse parameter will be removed.
    entityType: list[EntityType] | None = None
    language: Language | None = None
    allowFreeformNewLine: bool | None = None
    returnMatchType: bool | None = None
    view: ViewType | None = None
    mapcodes: list[MapCodeType] | None = None
    filter: FilterType | None = None


@dataclass(kw_only=True)
class ReverseGeocodeResponse(DataClassORJSONMixin):
    """Represents a reverse geocode response."""

    summary: Summary
    addresses: list[Addresses]


class RoadClassType(StrEnum):
    """Supported road class types."""

    FUNCTIONAL = "Functional"


class RoadClassTypeValue(StrEnum):
    """Represents a classification of roads."""

    MOTORWAY = "Motorway"
    TRUNK = "Trunk"
    PRIMARY = "Primary"
    SECONDARY = "Secondary"
    TERTIARY = "Tertiary"
    STREET = "Street"
    FERRY = "Ferry"
    OTHER = "Other"


@dataclass(kw_only=True)
class Roadclass(DataClassORJSONMixin):
    """Represents Roadclass."""

    type: RoadClassType
    values: list[RoadClassTypeValue]


@dataclass(kw_only=True)
class Route:
    """Represents an Route."""

    points: list[LatLon]


@dataclass(kw_only=True)
class SearchAlongRouteData(BasePostData):
    """Data for the post search along route API."""

    route: Points


@dataclass(kw_only=True)
class SearchAlongRouteParams(BaseParams):
    """Parameters for the post_search_along_route method."""

    typeahead: bool | None = None
    limit: int | None = None
    categorySet: list[str] | None = None
    brandSet: list[str] | None = None
    connectorSet: list[ConnectorType] | None = None
    minPowerKW: float | None = None
    maxPowerKW: float | None = None
    fuelSet: list[FuelType] | None = None
    vehicleTypeSet: list[VehicleType] | None = None
    view: ViewType | None = None
    detourOffset: bool | None = None
    sortBy: SortByType | None = None
    language: Language | None = None
    openingHours: OpeningHoursType | None = None
    spreadingMode: SpreadingMode | None = None
    mapcodes: list[MapCodeType] | None = None
    timeZone: str | None = None
    relatedPois: RelatedPoisType | None = None


@dataclass(kw_only=True)
class SearchParams(BaseParams):
    """Parameters for the get_search method."""

    typeahead: bool | None = None
    limit: int | None = None
    ofs: int | None = None
    countrySet: list[str] | None = None
    lat: float | None = None
    lon: float | None = None
    radius: int | None = None
    topLeft: str | None = None
    btmRight: str | None = None
    geoBias: str | None = None
    language: Language | None = None
    extendedPostalCodesFor: list[ExtendedPostalCodesForType] | None = None
    minFuzzyLevel: int | None = None
    maxFuzzyLevel: int | None = None
    idxSet: list[IdxSetType] | None = None
    categorySet: list[str] | None = None
    brandSet: list[str] | None = None
    connectorSet: list[ConnectorType] | None = None
    minPowerKW: float | None = None
    maxPowerKW: float | None = None
    fuelSet: list[FuelType] | None = None
    vehicleTypeSet: list[VehicleType] | None = None
    view: ViewType | None = None
    openingHours: OpeningHoursType | None = None
    timeZone: str | None = None
    mapcodes: list[MapCodeType] | None = None
    relatedPois: RelatedPoisType | None = None
    entityTypeSet: list[EntityType] | None = None


@dataclass(kw_only=True)
class SearchResponse(DataClassORJSONMixin):
    """Represents a Search response."""

    summary: Summary
    results: list[Result]


@dataclass(kw_only=True)
class Segment(DataClassORJSONMixin):
    """Represents an auto complete segment."""

    type: str
    value: str
    matches: Matches
    id: str | None = None
    matchedAlternativeName: str | None = None


class SortByType(StrEnum):
    """Supported sort by type."""

    DETOUR_TIME = "detourTime"
    DETOUR_OFFSET = "detourOffset"


class SpreadingMode(StrEnum):
    """Supported spreading mode types."""

    AUTO = "auto"


class StatusType(StrEnum):
    """Supported current types."""

    AVAILABLE = "Available"
    RESERVED = "Reserved"
    OCCUPIED = "Occupied"
    OUT_OF_SERVICE = "OutOfService"
    UNKNOWN = "Unknown"


@dataclass(kw_only=True)
class StructuredGeocodeParams(BaseParams):
    """Represents the parameters for get_structured_geocode."""

    limit: int | None = None
    ofs: int | None = None
    streetNumber: str | None = None
    streetName: str | None = None
    crossStreet: str | None = None
    municipality: str | None = None
    municipalitySubdivision: str | None = None
    countryTertiarySubdivision: str | None = None
    countrySecondarySubdivision: str | None = None
    countrySubdivision: str | None = None
    postalCode: str | None = None
    language: Language | None = None
    extendedPostalCodesFor: list[ExtendedPostalCodesForType] | None = None
    view: ViewType | None = None
    entityTypeSet: list[EntityType] | None = None


@dataclass(kw_only=True)
class Summary(DataClassORJSONMixin):
    """Represents a response Summary."""

    query: str | None = None
    queryTime: int | None = None
    queryType: QueryType | None = None  # Deprecated
    numResults: int
    offset: int | None = None
    totalResults: int | None = None
    fuzzyLevel: int | None = None
    geoBias: LatLon | None = None
    queryIntent: list[QueryIntent] | None = None


@dataclass(kw_only=True)
class Time(DataClassORJSONMixin):
    """Represents a Time."""

    date: date
    hour: int
    minute: int


@dataclass(kw_only=True)
class TimeRange(DataClassORJSONMixin):
    """Represents a TimeRange."""

    startTime: Time
    endTime: Time


@dataclass(kw_only=True)
class TimeZone(DataClassORJSONMixin):
    """Represents a TimeZone."""

    ianaId: str


class VehicleType(StrEnum):
    """Supported vehicle types."""

    CAR = "Car"
    TRUCK = "Truck"


@dataclass(kw_only=True)
class Viewport(DataClassORJSONMixin):
    """Represents a Viewport."""

    topLeftPoint: LatLon
    btmRightPoint: LatLon


@dataclass(kw_only=True)
class W3WQueryIntent(DataClassORJSONMixin):
    """Represents a W3WQueryIntent."""

    address: str



"""Batch Search API."""

from ..api import BaseApi, BaseParams
from ..places.models import (
    AsynchronousBatchDownloadParams,
    AsynchronousSynchronousBatchParams,
    BatchPostData,
    BatchResponse,
)


class BatchSearchApi(BaseApi):
    """Batch Search API.

    Batch Search sends batches of requests to supported endpoints with ease. You can call Batch Search APIs to run either asynchronously or
    synchronously. The Batch Search service consists of the following endpoints:

    For more information, see: https://developer.tomtom.com/batch-search-api/documentation/product-information/introduction
    """

    async def post_synchronous_batch(
        self,
        *,
        params: BaseParams | None = None,  # No extra params.
        data: BatchPostData,
    ) -> BatchResponse:
        """Post synchronous batch.

        For more information, see: https://developer.tomtom.com/batch-search-api/documentation/synchronous-batch

        Args:
            params (BaseParams, optional): Query parameters for the request. Defaults to None.
            data (BatchPostData): Data for the batch request.

        Returns:
            BatchResponse: The response object for the synchronous batch request.
        """
        response = await self.post(
            endpoint="/search/2/batch/sync.json",
            params=params,
            data=data,
        )

        return await response.deserialize(BatchResponse)

    async def post_asynchronous_batch_submission(
        self,
        *,
        params: AsynchronousSynchronousBatchParams | None = None,
        data: BatchPostData,
    ) -> str | None:
        """Post Asynchronous Batch Submission.

        For more information, see: https://developer.tomtom.com/batch-search-api/documentation/asynchronous-batch-submission

        Args:
            params (AsynchronousSynchronousBatchParams, optional): Query parameters for the request. Defaults to None.
            data (BatchPostData): Data for the batch request.

        Returns:
            str | None: The 'Location' header from the response, if available, otherwise None.
        """
        response = await self.post(
            endpoint="/search/2/batch.json",
            params=params,
            data=data,
        )

        return response.headers.get("Location", None)

    async def get_asynchronous_batch_download(
        self,
        *,
        batch_id: str,
        params: AsynchronousBatchDownloadParams | None = None,
    ) -> BatchResponse:
        """Fetches the result of an asynchronous batch download.

        For more information, see: https://developer.tomtom.com/batch-search-api/documentation/asynchronous-batch-download

        Args:
            batch_id (str): The ID of the batch to download.
            params (AsynchronousBatchDownloadParams, optional): Optional parameters for the download request. Defaults to None.

        Returns:
            BatchResponse: The response object representing the downloaded batch.
        """
        response = await self.get(
            endpoint=f"/search/2/batch/{batch_id}",
            params=params,
        )

        return await response.deserialize(BatchResponse)



"""Long Distance EV Routing API."""

from ..api import BaseApi
from ..models import LatLonList
from .models import CalculatedLongDistanceEVRouteResponse, CalculateLongDistanceEVRouteParams, CalculateLongDistanceEVRoutePostData


class LongDistanceEVRoutingApi(BaseApi):
    """Long Distance EV Routing API.

    The Long Distance EV Routing service endpoint calculates a route between a given origin and destination, passing through waypoints if they are
    specified. The route contains charging stops that have been added automatically based on the vehicle's consumption and charging model.

    For more information, see: https://developer.tomtom.com/long-distance-ev-routing-api/documentation/tomtom-maps/product-information/introduction
    """

    async def post_calculate_long_distance_ev_route(
        self,
        *,
        locations: LatLonList,
        params: CalculateLongDistanceEVRouteParams | None = None,
        data: CalculateLongDistanceEVRoutePostData,
    ) -> CalculatedLongDistanceEVRouteResponse:
        """Get long distance ev route.

        For more information, see:
        https://developer.tomtom.com/long-distance-ev-routing-api/documentation/tomtom-maps/product-information/introduction

        Args:
            locations (LatLonList): A list of locations to include in the route.
            params (CalculateLongDistanceEVRouteParams | None, optional): Additional parameters for the route calculation. Defaults to None.
            data (CalculateLongDistanceEVRoutePostData): Data specifying route details and constraints.

        Returns:
            CalculatedLongDistanceEVRouteResponse: The response containing the calculated long-distance EV route.
        """
        response = await self.post(
            endpoint=f"/routing/1/calculateLongDistanceEVRoute/{locations.to_colon_separated()}/json",
            params=params,
            data=data,
        )

        return await response.deserialize(CalculatedLongDistanceEVRouteResponse)



"""Premium Geocode API."""

from ..api import BaseApi
from .models import PremiumGeocodeParams, SearchResponse


class PremiumGeocodingApi(BaseApi):
    """Premium Geocoding API.

    The TomTom Premium Geocoding API is a forward geocoding service that returns highly accurate address coordinates along with expanded address
    features required for last-mile delivery, such as the closest parking points, floor numbering, and building entrances. This enables, for example,
    address verification, route planning, and front-door navigation. The highly accurate data that comes with premium geocoding allows delivery
    couriers to get to the customer's door much quicker compared to the standard single-point geocoded location available in regular geocoding. The
    Premium Geocoding API is currently only available in the USA.

    For more information, see: https://developer.tomtom.com/premium-geocoding-api/documentation/product-information/introduction
    """

    async def get_geocode(
        self,
        *,
        query: str,
        params: PremiumGeocodeParams | None = None,
    ) -> SearchResponse:
        """Get geocode.

        For more information, see: https://developer.tomtom.com/premium-geocoding-api/documentation/geocode

        Args:
            query (str): The query string representing the address or place to geocode.
            params (PremiumGeocodeParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            SearchResponse: Response containing search results.
        """
        response = await self.get(
            endpoint=f"/search/2/premiumGeocode/{query}.json",
            params=params,
        )

        return await response.deserialize(SearchResponse)



"""Models for the TomTom Automotive API."""
# pylint: disable=invalid-name, too-many-instance-attributes, too-many-lines

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum

from mashumaro.mixins.orjson import DataClassORJSONMixin

from ..api import BaseParams


@dataclass(kw_only=True)
class Current(DataClassORJSONMixin):
    """Represents a Current."""

    available: bool
    emptySpots: int
    availabilityTrend: str
    updatedAt: datetime


@dataclass(kw_only=True)
class Fuel(DataClassORJSONMixin):
    """Represents a Fuel."""

    type: FuelType
    price: list[Price]
    updatedAt: datetime


@dataclass(kw_only=True)
class FuelPricesResponse(DataClassORJSONMixin):
    """Represents a FuelPrices response."""

    fuelPrice: str
    fuels: list[Fuel]


@dataclass(kw_only=True)
class FuelPrizeParams(BaseParams):
    """Parameters for the get_fuel_prize method."""

    fuelPrice: str


class FuelType(StrEnum):
    """Supported fuel types."""

    BIODIESEL = "biodiesel"
    CNG = "cng"
    DIESEL = "diesel"
    DIESEL_PLUS = "dieselPlus"
    DIESEL_PLUS_WITHOUT_ADDITIVES = "dieselPlusWithoutAdditives"
    DIESEL_WITHOUT_ADDITIVES = "dieselWithoutAdditives"
    E100 = "e100"
    E80 = "e80"
    E85 = "e85"
    ETHANOL_WITHOUT_ADDITIVES = "ethanolWithoutAdditives"
    LPG = "lpg"
    REGULAR = "regular"
    SP100 = "sp100"
    SP91 = "sp91"
    SP91_E10 = "sp91_e10"
    SP92 = "sp92"
    SP92_PLUS = "sp92Plus"
    SP93 = "sp93"
    SP95 = "sp95"
    SP95_E10 = "sp95_e10"
    SP95_PLUS = "sp95Plus"
    SP95_WITHOUT_ADDITIVES = "sp95WithoutAdditives"
    SP97 = "sp97"
    SP98 = "sp98"
    SP98_PLUS = "sp98Plus"
    SP99 = "sp99"


@dataclass(kw_only=True)
class ParkingAvailabilityParams(BaseParams):
    """Parameters for the get_parking_availability method."""

    parkingAvailability: str


@dataclass(kw_only=True)
class ParkingAvailabilityResponse(DataClassORJSONMixin):
    """Represents a ParkingAvailability response."""

    parkingAvailability: str
    statuses: list[Status]


@dataclass(kw_only=True)
class Price(DataClassORJSONMixin):
    """Represents a Price."""

    value: float
    currency: str
    currencySymbol: str
    volumeUnit: str


@dataclass(kw_only=True)
class Status(DataClassORJSONMixin):
    """Represents a Status."""

    current: Current



"""Auto Stream API."""

from ..api import BaseApi


class AutoStreamApi(BaseApi):
    """AutoStream API.

    AutoStream is a map data delivery platform, optimized for on-demand and over-the-air cloud-to-device and cloud-to-cloud data streaming.

    For more information, see: https://developer.tomtom.com/autostream-sdk/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError

    # There are no methods defined on the developer portal.



"""Client for the TomTom API."""

from __future__ import annotations

import logging
import socket
import uuid
from dataclasses import dataclass, field
from importlib import metadata
from typing import Any, Literal, TypeVar

import orjson
from aiohttp import ClientResponse, ClientTimeout
from aiohttp.client import (
    ClientConnectionError,
    ClientError,
    ClientResponseError,
    ClientSession,
)
from aiohttp.hdrs import ACCEPT_ENCODING, CONTENT_TYPE, USER_AGENT
from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .const import TOMTOM_HEADER_PREFIX, TRACKING_ID_HEADER
from .exceptions import (
    TomTomAPIClientError,
    TomTomAPIConnectionError,
    TomTomAPIError,
    TomTomAPIRequestTimeout,
    TomTomAPIServerError,
)
from .utils import serialize_bool, serialize_list

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class BaseParams(DataClassDictMixin):
    """Base class for any params data class.

    Attributes:
        key (str | None): The api key attribute, defaults to None, can override the key from ApiOptions.
    """

    key: str | None = None

    def __post_serialize__(self, d: dict[Any, Any]) -> dict[str, str]:
        """Removes keys with None values from the serialized dictionary.

        Args:
            d: The dictionary to be processed.

        Returns:
            A new dictionary without keys that have a value of None.
        """
        return {k: v for k, v in d.items() if v is not None}

    class Config(BaseConfig):  # pylint: disable=too-few-public-methods
        """Config for the BaseParams.

        Not setting omit_none=True, because that runs before serialization, while in serialization empty lists are set to None.
        Manually omitting None values in __post_serialize__ to fix this.
        """

        serialization_strategy = {
            bool: {
                "serialize": serialize_bool,
            },
            float: {
                "serialize": str,
            },
            int: {
                "serialize": str,
            },
            list: {
                "serialize": serialize_list,
            },
        }


@dataclass(kw_only=True)
class BasePostData(DataClassDictMixin):
    """Base class for any post data class.

    Attributes:
        DataClassDictMixin: Mixin for converting data classes to dictionaries.
    """


class Response:
    """Response class for the TomTom API.

    Args:
        response: The aiohttp ClientResponse object.

    Methods:
        deserialize(model: type[T]) -> T: Deserialize the response to the given model.
        dict() -> dict: Deserialize the response to a dictionary.
        text() -> str: Return the response as text.
        bytes() -> bytes: Return the response as bytes.
    """

    T = TypeVar("T", bound=DataClassORJSONMixin)

    def __init__(self, response: ClientResponse):
        """Initialize the Response object.

        Args:
            response: The aiohttp ClientResponse object.
        """
        self._response = response
        self.headers: dict[str, str] = dict(response.headers)
        self.status = response.status

    async def deserialize(self, model: type[T]) -> T:
        """Deserialize the response to the given model.

        Args:
            model: The model class to deserialize the response to.

        Returns:
            An instance of the given model class.

        Raises:
            Exception: If the deserialization fails.
        """
        logger.info("Deserializing response to %s", model)
        try:
            text = await self._response.text()
            return model.from_json(text)
        except Exception as e:
            logger.error("Failed to deserialize response: %s", e)
            raise

    async def dict(self) -> dict:
        """Deserialize the response to a dictionary.

        Returns:
            A dictionary representation of the response.

        Raises:
            orjson.JSONDecodeError: If the response is not valid JSON.
        """
        logger.info("Deserializing response to dictionary")
        try:
            text = await self._response.text()
            return orjson.loads(text)  # pylint: disable=maybe-no-member
        except orjson.JSONDecodeError as e:  # pylint: disable=maybe-no-member
            logger.error("Failed to decode JSON response: %s", e)
            raise

    async def text(self) -> str:
        """Return the response as text.

        Returns:
            The response as a string.
        """
        logger.info("Returning response as text")
        return await self._response.text()

    async def bytes(self) -> bytes:
        """Return the response as bytes.

        Returns:
            The response as a bytes object.
        """
        logger.info("Returning response as bytes")
        return await self._response.read()


@dataclass(kw_only=True)
class ApiOptions:
    """Options to configure the TomTom API client.

    Attributes:
        api_key: str
            An API key valid for the requested service.
        base_url: str
            The base URL for the TomTom API. Default is "https://api.tomtom.com".
        gzip_compression: bool, optional
            Enables response compression. Default is False.
        timeout: ClientTimeout, optional
            The timeout object for the request. Default is ClientTimeout(total=10).
        tracking_id: bool, optional
            Specifies an identifier for each request. Default is False.
    """

    api_key: str
    base_url: str = "https://api.tomtom.com"
    gzip_compression: bool = False
    timeout: ClientTimeout = field(default_factory=lambda: ClientTimeout(total=10))
    tracking_id: bool = False


class BaseApi:
    """Client for the TomTom API.

    Attributes:
        options : ApiOptions
            The options for the client.
    """

    _version: str = metadata.version(__package__)

    def __init__(
        self,
        options: ApiOptions,
        session: ClientSession | None = None,
    ):
        """Initializes the BaseApi object.

        Args:
            options: ApiOptions
                The options for the client.
            session: ClientSession, optional
                The client session to use for requests. If not provided, a new session is created.
        """
        self.options = options
        self.session = ClientSession(options.base_url, timeout=options.timeout) if session is None else session

        self._default_headers: dict[str, str] = {
            CONTENT_TYPE: "application/json",
            USER_AGENT: f"TomTomApiPython/{self._version}",
        }

        self._default_params: dict[str, str] = {
            "key": options.api_key,
        }

    async def _request(  # pylint: disable=too-many-arguments
        self,
        method: Literal["DELETE", "GET", "POST", "PUT"],
        endpoint: str,
        *,
        headers: dict[str, str] | None = None,
        params: BaseParams | None = None,
        data: BasePostData | None = None,
    ) -> Response:
        """Make a request to the TomTom API.

        Args:
            method: Literal["DELETE", "GET", "POST", "PUT"]
                The HTTP method for the request.
            endpoint: str
                The endpoint to send the request to.
            headers: dict[str, str] | None, optional
                The headers for the request.
            params: BaseParams | None, optional
                The parameters for the request.
            data: BasePostData | None, optional
                The data to be sent in the request body.

        Returns:
            Response
                The response object from the API.

        Raises:
            TomTomAPIRequestTimeout: If a timeout occurs while connecting to the API.
            TomTomAPIConnectionError: If a connection error occurs.
            TomTomAPIClientError: If a client-side error (4xx) occurs.
            TomTomAPIServerError: If a server-side error (5xx) occurs.
            TomTomAPIError: For other errors raised by the TomTom SDK.
        """
        request_params = {
            **self._default_params,
            **(params.to_dict() if params else {}),
        }
        request_headers = {**self._default_headers, **(headers if headers else {})}
        request_data = data.to_dict() if data else None

        if self.options.gzip_compression:
            request_headers[ACCEPT_ENCODING] = "gzip"
        if self.options.tracking_id:
            tracking_id = str(uuid.uuid4())
            request_headers[TRACKING_ID_HEADER] = tracking_id
        else:
            tracking_id = "not tracked"

        logger.info("%s %s (%s)", method, endpoint, tracking_id)

        try:
            response = await self.session.request(
                method,
                endpoint,
                params=request_params,
                json=request_data,
                headers=request_headers,
            )

            logger.info("%s %s returns: %s", method, endpoint, response.status)

            # Log TomTom and the tracking id headers
            for header, value in response.headers.items():
                if header.lower().startswith(TOMTOM_HEADER_PREFIX) or header.lower() == TRACKING_ID_HEADER.lower():
                    logger.info("Response header %s: %s", header, value)

            response.raise_for_status()

        except TimeoutError as exception:
            msg = "Timeout occurred while connecting to the API"
            raise TomTomAPIRequestTimeout(msg) from exception
        except ClientConnectionError as exception:
            msg = "Connection error"
            raise TomTomAPIConnectionError(msg) from exception
        except ClientResponseError as exception:
            if 400 <= exception.status < 500:
                msg = "Client error"
                raise TomTomAPIClientError(msg) from exception
            if exception.status >= 500:
                msg = "Server error"
                raise TomTomAPIServerError(msg) from exception
            msg = "Response error"
            raise TomTomAPIError(msg) from exception
        except (
            ClientError,
            socket.gaierror,
        ) as exception:
            msg = "Error occurred while communicating with the API"
            raise TomTomAPIConnectionError(exception) from exception

        return Response(response)

    async def delete(
        self,
        endpoint: str,
        *,
        headers: dict[str, str] | None = None,
        params: BaseParams | None = None,
    ) -> Response:
        """Make a DELETE request.

        Args:
            endpoint: str
                The endpoint to send the DELETE request to.
            headers: dict[str, str] | None, optional
                The headers for the request.
            params: BaseParams | None, optional
                The parameters for the request.

        Returns:
            Response
                The response object from the API.
        """
        return await self._request(
            "DELETE",
            endpoint,
            headers=headers,
            params=params,
        )

    async def get(
        self,
        endpoint: str,
        *,
        headers: dict[str, str] | None = None,
        params: BaseParams | None = None,
    ) -> Response:
        """Make a GET request.

        Args:
            endpoint: str
                The endpoint to send the GET request to.
            headers: dict[str, str] | None, optional
                The headers for the request.
            params: BaseParams | None, optional
                The parameters for the request.

        Returns:
            Response
                The response object from the API.
        """
        return await self._request(
            "GET",
            endpoint,
            headers=headers,
            params=params,
        )

    async def post(  # pylint: disable=too-many-arguments
        self,
        endpoint: str,
        *,
        headers: dict[str, str] | None = None,
        params: BaseParams | None = None,
        data: BasePostData,
    ) -> Response:
        """Make a POST request.

        Args:
            endpoint: str
                The endpoint to send the POST request to.
            headers: dict[str, str] | None, optional
                The headers for the request.
            params: BaseParams | None, optional
                The parameters for the request.
            data: BasePostData
                The data to be sent in the request body.

        Returns:
            Response
                The response object from the API.
        """
        return await self._request(
            "POST",
            endpoint,
            headers=headers,
            params=params,
            data=data,
        )

    async def put(  # pylint: disable=too-many-arguments
        self,
        endpoint: str,
        *,
        headers: dict[str, str] | None = None,
        params: BaseParams | None = None,
        data: BasePostData,
    ) -> Response:
        """Make a PUT request.

        Args:
            endpoint: str
                The endpoint to send the PUT request to.
            headers: dict[str, str] | None, optional
                The headers for the request.
            params: BaseParams | None, optional
                The parameters for the request.
            data: BasePostData
                The data to be sent in the request body.

        Returns:
            Response
                The response object from the API.
        """
        return await self._request(
            "PUT",
            endpoint,
            headers=headers,
            params=params,
            data=data,
        )

    async def __aenter__(self):
        """Enter the runtime context related to this object.

        The session used to make requests is created upon entering the context and closed upon exiting.

        Returns:
            self
        """
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context related to this object.

        The session used for making requests is closed upon exiting the context.

        Args:
            exc_type: The type of the exception raised in the context.
            exc_val: The value of the exception raised in the context.
            exc_tb: The traceback of the exception raised in the context.
        """
        if self.session:
            await self.session.close()  # Close the session when exiting the context
            self.session = None

    async def close(self):
        """Close the session.

        Manually closes the session. If the session is not closed, it will be closed when exiting the context.

        Note:
            Does not raise an exception if the session is already closed.
        """
        if self.session:
            await self.session.close()  # Close the session if manually closing
            self.session = None



"""Generic models."""

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum, StrEnum

from mashumaro.mixins.orjson import DataClassORJSONMixin


class AdrCategoryType(StrEnum):
    """Supported ADR category types."""

    B = "B"
    C = "C"
    D = "D"
    E = "E"


class Language(StrEnum):
    """Supported languages for the Places API.

    For more information, see: https://developer.tomtom.com/search-api/documentation/product-information/supported-languages.
    """

    NGT = "NGT"  # Neutral Ground Truth, Official languages for all regions in local scripts if available.
    NGT_LATN = "NGT-Latn"  # Neutral Ground Truth - Latin exonyms, Latin script will be used if available.
    AF_ZA = "af-ZA"  # Afrikaans
    AR = "ar"  # Arabic
    EU_ES = "eu-ES"  # Basque
    BG_BG = "bg-BG"  # Bulgarian
    CA_ES = "ca-ES"  # Catalan (Spain)
    ZH_CN = "zh-CN"  # Chinese (PRC)
    ZH_TW = "zh-TW"  # Chinese (Taiwan)
    CS_CZ = "cs-CZ"  # Czech
    DA_DK = "da-DK"  # Danish
    NL_BE = "nl-BE"  # Dutch (Belgium)
    NL_NL = "nl-NL"  # Dutch
    EN_AU = "en-AU"  # English (Australia)
    EN_NZ = "en-NZ"  # English (New Zealand)
    EN_GB = "en-GB"  # English (Great Britain)
    EN_US = "en-US"  # English (USA)
    ET_EE = "et-EE"  # Estonian
    FI_FI = "fi-FI"  # Finnish
    FR_CA = "fr-CA"  # French (Canada)
    FR_FR = "fr-FR"  # French
    GL_ES = "gl-ES"  # Galician
    DE_DE = "de-DE"  # German
    EL_GR = "el-GR"  # Greek
    HR_HR = "hr-HR"  # Croatian
    HE_IL = "he-IL"  # Hebrew
    HU_HU = "hu-HU"  # Hungarian
    ID_ID = "id-ID"  # Indonesian
    IT_IT = "it-IT"  # Italian
    KK_KZ = "kk-KZ"  # Kazakh
    KO_KR = "ko-KR"  # Korean written in the Hangul script
    KO_LATN_KR = "ko-Latn-KR"  # Korean written in the Latin script
    KO_KORE_KR = "ko-Kore-KR"  # Korean written in the Hangul script
    LV_LV = "lv-LV"  # Latvian
    LT_LT = "lt-LT"  # Lithuanian
    MS_MY = "ms-MY"  # Malay
    NO_NO = "no-NO"  # Norwegian
    NB_NO = "nb-NO"  # Norwegian
    PL_PL = "pl-PL"  # Polish
    PT_BR = "pt-BR"  # Portuguese (Brazil)
    PT_PT = "pt-PT"  # Portuguese (Portugal)
    RO_RO = "ro-RO"  # Romanian
    RU_RU = "ru-RU"  # Russian written in the Cyrillic script
    RU_LATN_RU = "ru-Latn-RU"  # Russian written in the Latin script
    RU_CYRL_RU = "ru-Cyrl-RU"  # Russian written in the Cyrillic script
    SR_RS = "sr-RS"  # Serbian
    SK_SK = "sk-SK"  # Slovak
    SL_SI = "sl-SI"  # Slovenian
    ES_ES = "es-ES"  # Castilian Spanish
    ES_419 = "es-419"  # Latin American Spanish
    SV_SE = "sv-SE"  # Swedish
    TH_TH = "th-TH"  # Thai
    TR_TR = "tr-TR"  # Turkish
    UK_UA = "uk-UA"  # Ukrainian
    VI_VN = "vi-VN"  # Vietnamese


@dataclass(kw_only=True)
class LatLon(DataClassORJSONMixin):
    """Point location with abbreviated names.

    Attributes:
        lat (float): The latitude of the point.
        lon (float): The longitude of the point.

    """

    lat: float
    lon: float

    def to_comma_separated(self) -> str:
        """Return the lat, lon as a comma-separated string.

        Returns:
            str: The lat, lon as a comma-separated string.

        """
        return ",".join(map(str, [self.lat, self.lon]))


@dataclass(kw_only=True)
class LatLonList(DataClassORJSONMixin):
    """Dataclass to handle a list of LatLon objects.

    Attributes:
        locations (list[LatLon]): The list of LatLon objects.

    """

    locations: list[LatLon]

    def to_colon_separated(self) -> str:
        """Return the list of lat, lon as a colon-separated string.

        Returns:
            str: The list of lat, lon as a colon-separated string.

        """
        return ":".join([loc.to_comma_separated() for loc in self.locations])


@dataclass(kw_only=True)
class LatitudeLongitude(DataClassORJSONMixin):
    """Point location with full names.

    Attributes:
        latitude (float): The latitude of the point.
        longitude (float): The longitude of the point.

    """

    latitude: float
    longitude: float


@dataclass(kw_only=True)
class MapTile:
    """Map tile representation.

    Attributes:
        x (int): The x coordinate.
        y (int): The y coordinate.
        zoom (int): The zoom level

    """

    x: int
    y: int
    zoom: int


class TileSizeType(IntEnum):
    """Supported tile sizes."""

    SIZE_256 = 256
    SIZE_512 = 512


class TravelModeType(StrEnum):
    """Supported travel mode types."""

    CAR = "car"
    TRUCK = "truck"
    TAXI = "taxi"
    BUS = "bus"
    VAN = "van"
    MOTORCYCLE = "motorcycle"
    BICYCLE = "bicycle"
    PEDESTRIAN = "pedestrian"
    OTHER = "other"


class ViewType(StrEnum):
    """Geopolitical View.

    The context used to resolve the handling of disputed territories. Views include Unified, along with AR IL, IN, MA, PK, RU, TR, and CN which are
    respectively tailored for Argentina, Israel, India, Morocco, Pakistan, Russia, Turkey, and China.
    """

    UNIFIED = "Unified"
    AR = "AR"  # Argentina
    IL = "IL"  # Israel
    IN = "IN"  # India
    KR = "KR"  # Korea
    MA = "MA"  # Morocco
    PK = "PK"  # Pakistan
    RU = "RU"  # Russia
    TR = "TR"  # Turkey
    CN = "CN"  # China



"""Fuel Prices API."""

from ..api import BaseApi
from ..automotive.models import FuelPricesResponse, FuelPrizeParams


class FuelPricesApi(BaseApi):
    """Fuel Prices API.

    For more information, see: https://developer.tomtom.com/fuel-prices-api/documentation/product-information/introduction
    """

    async def get_fuel_prize(
        self,
        *,
        params: FuelPrizeParams | None = None,
    ) -> FuelPricesResponse:
        """Get fuel prize.

        For more information, see: https://developer.tomtom.com/fuel-prices-api/documentation/fuel-prices-api/fuel-price

        Args:
            params (FuelPrizeParams | None, optional): Additional parameters to filter the fuel prices query. Defaults to None.

        Returns:
            FuelPricesResponse: The response containing the fuel prices data.
        """
        response = await self.get(
            endpoint="/search/2/fuelPrice.json",
            params=params,
        )

        return await response.deserialize(FuelPricesResponse)



"""Automotive APIs."""

from .autostream import AutoStreamApi
from .fuel_prices import FuelPricesApi
from .parking_availability import ParkingAvailabilityApi

__all__ = [
    "AutoStreamApi",
    "FuelPricesApi",
    "ParkingAvailabilityApi",
]



"""Map Display API."""

from ..api import BaseApi, BaseParams
from .models import (
    LayerType,
    LayerTypeWithPoiType,
    MapServiceCopyrightsResponse,
    MapTileParams,
    MapTileV1Params,
    MapTileV2Params,
    StaticImageParams,
    StyleType,
    TileFormatType,
)


class MapDisplayApi(BaseApi):
    """Map Display API.

    The Map Display API is a suite of web services designed for developers to create web and mobile applications around mapping.

    For more information, see: https://developer.tomtom.com/map-display-api/documentation/product-information/introduction
    """

    async def get_map_tile(  # pylint: disable=too-many-arguments
        self,
        *,
        layer: LayerType,
        style: StyleType,
        x: int,
        y: int,
        zoom: int,
        image_format: TileFormatType,
        params: MapTileParams | None = None,
    ) -> bytes:
        """Get map tile.

        For more information, see: https://developer.tomtom.com/map-display-api/documentation/raster/map-tile

        Args:
            layer (LayerType): The type of layer for the map tile.
            style (StyleType): The style of the map tile.
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            zoom (int): The zoom level of the tile.
            image_format (TileFormatType): The format of the image.
            params (MapTileParams | None, optional): Additional parameters for the map tile. Defaults to None.

        Returns:
            bytes: The map tile image data.
        """
        response = await self.get(
            endpoint=f"/map/1/tile/{layer}/{style}/{zoom}/{x}/{y}.{image_format}",
            params=params,
        )

        return await response.bytes()

    async def get_satellite_tile(  # pylint: disable=too-many-arguments
        self,
        *,
        x: int,
        y: int,
        zoom: int,
        image_format: TileFormatType,
        params: BaseParams | None = None,  # No extra params.
    ) -> bytes:
        """Get satellite tile.

        For more information, see: https://developer.tomtom.com/map-display-api/documentation/raster/satellite-tile

        Args:
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            zoom (int): The zoom level of the tile.
            image_format (TileFormatType): The format of the image.
            params (BaseParams | None, optional): Additional parameters for the tile. Defaults to None.

        Returns:
            bytes: The satellite tile image data.
        """
        response = await self.get(
            endpoint=f"/map/1/tile/sat/main/{zoom}/{x}/{y}.{image_format}",
            params=params,
        )

        return await response.bytes()

    async def get_hillshade_tile(  # pylint: disable=too-many-arguments
        self,
        *,
        x: int,
        y: int,
        zoom: int,
        image_format: TileFormatType,
        params: BaseParams | None = None,  # No extra params.
    ) -> bytes:
        """Get hillshade tile.

        For more information, see: https://developer.tomtom.com/map-display-api/documentation/raster/hillshade-tile

        Args:
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            zoom (int): The zoom level of the tile.
            image_format (TileFormatType): The format of the image.
            params (BaseParams | None, optional): Additional parameters for the hillshade tile. Defaults to None.

        Returns:
            bytes: The hillshade tile image data.
        """
        response = await self.get(
            endpoint=f"/map/1/tile/hill/main/{zoom}/{x}/{y}.{image_format}",
            params=params,
        )

        return await response.bytes()

    async def get_static_image(  # pylint: disable=too-many-arguments
        self,
        *,
        params: StaticImageParams | None = None,
    ) -> bytes:
        """Get static image.

        For more information, see: https://developer.tomtom.com/map-display-api/documentation/raster/static-image

        Args:
            layer (LayerType): The type of layer for the map tile.
            style (StyleType): The style of the map tile.
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            zoom (int): The zoom level of the tile.
            image_format (TileFormatType): The format of the image.
            params (MapTileParams | None, optional): Additional parameters for the map tile. Defaults to None.

        Returns:
            bytes: The map tile image data.
        """
        response = await self.get(
            endpoint="/map/1/staticimage",
            params=params,
        )

        return await response.bytes()

    async def get_tile_v1(  # pylint: disable=too-many-arguments
        self,
        *,
        layer: LayerTypeWithPoiType,
        x: int,
        y: int,
        zoom: int,
        params: MapTileV1Params | None = None,
    ) -> bytes:
        """Get tile version 1.

        For more information, see: https://developer.tomtom.com/map-display-api/documentation/vector/tile

        Args:
            layer (LayerTypeWithPoiType): The type of layer for the map tile.
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            zoom (int): The zoom level of the tile.
            params (MapTileV1Params | None, optional): Additional parameters for the map tile. Defaults to None.

        Returns:
            bytes: The map tile data in bytes format.
        """
        response = await self.get(
            endpoint=f"/map/1/tile/{layer}/main/{zoom}/{x}/{y}.pbf",
            params=params,
        )

        return await response.bytes()

    async def get_tile_v2(  # pylint: disable=too-many-arguments
        self,
        *,
        layer: LayerTypeWithPoiType,
        x: int,
        y: int,
        zoom: int,
        params: MapTileV2Params | None = None,
    ) -> bytes:
        """Get tile version 2.

        For more information, see: https://developer.tomtom.com/map-display-api/documentation/vector/tile-v2

        Args:
            layer (LayerTypeWithPoiType): The type of layer for the map tile.
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            zoom (int): The zoom level of the tile.
            params (MapTileV2Params | None, optional): Additional parameters for the map tile. Defaults to None.

        Returns:
            bytes: The map tile data in bytes format.
        """
        response = await self.get(
            endpoint=f"/map/1/tile/{layer}/{zoom}/{x}/{y}.pbf",
            params=params,
        )

        return await response.bytes()

    async def get_map_copyrights(
        self,
        *,
        params: BaseParams | None = None,  # No extra params.
    ) -> str:
        """Get map copyrights.

        For more information, see: https://developer.tomtom.com/map-display-api/documentation/copyrights

        Args:
            params (BaseParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            str: The copyright information as a string.
        """
        response = await self.get(
            endpoint="/map/2/copyrights",
            params=params,
        )

        return await response.text()

    async def get_map_service_copyrights(
        self,
        *,
        params: BaseParams | None = None,  # No extra params.
    ) -> MapServiceCopyrightsResponse:
        """Get map service copyrights.

        For more information, see: https://developer.tomtom.com/map-display-api/documentation/copyrights

        Args:
            params (BaseParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            MapServiceCopyrightsResponse: The copyright information response object.
        """
        response = await self.get(
            endpoint="/map/2/copyrights/caption.json",
            params=params,
        )

        return await response.deserialize(MapServiceCopyrightsResponse)



"""Traffic API."""

from tomtom_apis.utils import serialize_list

from ..api import BaseApi, BaseParams
from .models import (
    BBoxParam,
    BoudingBoxParam,
    FlowSegmentDataParams,
    FlowSegmentDataResponse,
    FlowStyleType,
    FlowType,
    IncidentDetailsParams,
    IncidentDetailsPostData,
    IncidentDetailsResponse,
    IncidentStyleType,
    IncidentTileFormatType,
    IncidentViewportResponse,
    RasterFlowTilesParams,
    RasterIncidentTilesParams,
    VectorFlowTilesParams,
    VectorIncidentTilesParams,
)


class TrafficApi(BaseApi):
    """Traffic API.

    The Traffic API is a suite of web services designed for developers to create web and mobile applications around real-time traffic. These web
    services can be used via RESTful APIs. The TomTom Traffic team offers a wide range of solutions to enable you to get the most out of your
    applications. Make use of the real-time traffic products or the historical traffic analytics to create applications and analysis that fits the
    needs of your end-users.

    For more information, see: https://developer.tomtom.com/traffic-api/documentation/product-information/introduction
    """

    async def get_incident_details(
        self,
        *,
        bbox: BBoxParam | None = None,
        ids: list[str] | None = None,
        params: IncidentDetailsParams | None = None,
    ) -> IncidentDetailsResponse:
        """Get incident details.

        For more information, see: https://developer.tomtom.com/traffic-api/documentation/traffic-incidents/incident-details

        Args:
            bbox (BBoxParam | None, optional): The bounding box to filter incidents by location. Defaults to None.
            ids (list[str] | None, optional): A list of incident IDs to retrieve details for. Defaults to None.
            params (IncidentDetailsParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            IncidentDetailsResponse: The response containing details of the traffic incidents.

        Raises:
            ValueError: If both `bbox` and `ids` are provided or neither is provided.
        """
        if bbox and not ids:
            mutually_exclusive_parameters = f"bbox={bbox.to_comma_separated()}"
        elif ids and not bbox:
            mutually_exclusive_parameters = f"ids={serialize_list(list(ids))}"
        else:
            raise ValueError("bbox and ids are mutually exclusive parameters")

        response = await self.get(
            endpoint=f"/traffic/services/5/incidentDetails?{mutually_exclusive_parameters}",
            params=params,
        )

        return await response.deserialize(IncidentDetailsResponse)

    async def post_incident_details(
        self,
        *,
        params: IncidentDetailsParams | None = None,
        data: IncidentDetailsPostData,
    ) -> IncidentDetailsResponse:
        """Post incident details.

        For more information, see: https://developer.tomtom.com/traffic-api/documentation/traffic-incidents/incident-details

        Args:
            params (IncidentDetailsParams | None, optional): Additional parameters for the request. Defaults to None.
            data (IncidentDetailsPostData): Data containing the incident details to be posted.

        Returns:
            IncidentDetailsResponse: The response containing the result of the posted incident details.
        """
        response = await self.post(
            endpoint="/traffic/services/5/incidentDetails",
            params=params,
            data=data,
        )

        return await response.deserialize(IncidentDetailsResponse)

    async def get_incident_viewport(  # pylint: disable=too-many-arguments
        self,
        *,
        bounding_box: BoudingBoxParam,
        bounding_zoom: int,
        overview_box: BoudingBoxParam,
        overview_zoom: int,
        copyright_information: bool,
        params: BaseParams | None = None,  # No extra params.
    ) -> IncidentViewportResponse:
        """Get incident viewport.

        For more information, see: https://developer.tomtom.com/traffic-api/documentation/traffic-incidents/incident-viewport

        Args:
            bounding_box (BoudingBoxParam): The bounding box defining the primary viewport for incidents.
            bounding_zoom (int): The zoom level for the primary viewport.
            overview_box (BoudingBoxParam): The bounding box defining the overview viewport.
            overview_zoom (int): The zoom level for the overview viewport.
            copyright_information (bool): Flag to include copyright information in the response.
            params (BaseParams | None, optional): Optional parameters for the request. Defaults to None.

        Returns:
            IncidentViewportResponse: The response containing details of traffic incidents within the defined viewport.
        """
        response = await self.get(
            endpoint=(
                f"/traffic/services/4/incidentViewport/{bounding_box.to_comma_separated()}/"
                f"{bounding_zoom}/{overview_box.to_comma_separated()}/"
                f"{overview_zoom}/{copyright_information}/json"
            ),
            params=params,
        )

        return await response.deserialize(IncidentViewportResponse)

    async def get_raster_incident_tile(  # pylint: disable=too-many-arguments
        self,
        *,
        style: IncidentStyleType,
        x: int,
        y: int,
        zoom: int,
        image_format: IncidentTileFormatType,
        params: RasterIncidentTilesParams | None = None,
    ) -> bytes:
        """Get raster incident tile.

        For more information, see: https://developer.tomtom.com/traffic-api/documentation/traffic-incidents/raster-incident-tiles

        Args:
            style (IncidentStyleType): The style of the incident tile (e.g., default or custom style).
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            zoom (int): The zoom level of the tile.
            image_format (IncidentTileFormatType): The format of the image.
            params (RasterIncidentTilesParams | None, optional): Optional parameters for the request. Defaults to None.

        Returns:
            bytes: The raster image tile in PNG format.
        """
        response = await self.get(
            endpoint=f"/traffic/map/4/tile/incidents/{style}/{zoom}/{x}/{y}.{image_format}",
            params=params,
        )

        return await response.bytes()

    async def get_vector_incident_tile(
        self,
        *,
        x: int,
        y: int,
        zoom: int,
        params: VectorIncidentTilesParams | None = None,
    ) -> bytes:
        """Get vector incident tile.

        For more information, see: https://developer.tomtom.com/traffic-api/documentation/traffic-incidents/vector-incident-tiles

        Args:
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            zoom (int): The zoom level of the tile.
            params (VectorIncidentTilesParams | None, optional): Optional parameters for the request. Defaults to None.

        Returns:
            bytes: The vector tile in PBF format.
        """
        response = await self.get(
            endpoint=f"/traffic/map/4/tile/incidents/{zoom}/{x}/{y}.pbf",
            params=params,
        )

        return await response.bytes()

    async def get_flow_segment_data(
        self,
        *,
        style: FlowStyleType,
        zoom: int,
        point: str,
        params: FlowSegmentDataParams | None = None,
    ) -> FlowSegmentDataResponse:
        """Get flow segment data.

        For more information, see: https://developer.tomtom.com/traffic-api/documentation/traffic-flow/flow-segment-data

        Args:
            style (FlowStyleType): The style of the flow segment data (e.g., default or custom style).
            zoom (int): The zoom level for the flow segment data.
            point (str): The coordinates of the point of interest, specified as a string.
            params (FlowSegmentDataParams | None, optional): Optional parameters for the request. Defaults to None.

        Returns:
            FlowSegmentDataResponse: The response containing traffic flow segment data.
        """
        response = await self.get(
            endpoint=f"/traffic/services/4/flowSegmentData/{style}/{zoom}/json?point={point}",
            params=params,
        )

        return await response.deserialize(FlowSegmentDataResponse)

    async def get_raster_flow_tiles(  # pylint: disable=too-many-arguments
        self,
        *,
        style: FlowStyleType,
        zoom: int,
        x: int,
        y: int,
        params: RasterFlowTilesParams | None = None,
    ) -> bytes:
        """Get raster flow tiles.

        For more information, see: https://developer.tomtom.com/traffic-api/documentation/traffic-flow/raster-flow-tiles

        Args:
            style (FlowStyleType): The style of the flow tile (e.g., default or custom style).
            zoom (int): The zoom level of the tile.
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            params (RasterFlowTilesParams | None, optional): Optional parameters for the request. Defaults to None.

        Returns:
            bytes: The raster image tile in PNG format.
        """
        response = await self.get(
            endpoint=f"/traffic/map/4/tile/flow/{style}/{zoom}/{x}/{y}.png",
            params=params,
        )

        return await response.bytes()

    async def get_vector_flow_tiles(  # pylint: disable=too-many-arguments
        self,
        *,
        flow_type: FlowType,
        zoom: int,
        x: int,
        y: int,
        params: VectorFlowTilesParams | None = None,
    ) -> bytes:
        """Get vector flow tiles.

        For more information, see: https://developer.tomtom.com/traffic-api/documentation/traffic-flow/vector-flow-tiles

        Args:
            flow_type (FlowType): The type of flow data to retrieve (e.g., current or historical flow).
            zoom (int): The zoom level of the tile.
            x (int): The x-coordinate of the tile.
            y (int): The y-coordinate of the tile.
            params (VectorFlowTilesParams | None, optional): Optional parameters for the request. Defaults to None.

        Returns:
            bytes: The vector tile in PBF format.
        """
        response = await self.get(
            endpoint=f"/traffic/map/4/tile/flow/{flow_type}/{zoom}/{x}/{y}.pbf",
            params=params,
        )

        return await response.bytes()



"""Exceptions for the TomTom API."""


class TomTomAPIError(Exception):
    """Base exception for all errors raised by the TomTom SDK."""


class TomTomAPIClientError(TomTomAPIError):
    """Exception raised for client-side errors (4xx)."""


class TomTomAPIServerError(TomTomAPIError):
    """Exception raised for server-side errors (5xx)."""


class TomTomAPIConnectionError(TomTomAPIError):
    """Exception raised for connection errors."""


class TomTomAPIRequestTimeout(TomTomAPIError):
    """Exception raised for request timeouts."""



"""Matrix Routing V2 API."""

from ..api import BaseApi


class MatrixRoutingApiV2(BaseApi):
    """Matrix Routing V2 API.

    The Matrix Routing v2 API allows clients to calculate a matrix of route summaries for a set of routes defined with origin and destination
    locations. For every given origin, this service calculates the cost of routing from that origin to every given destination. The set of origins
    and the set of destinations can be thought of as the column and row headers of a table, while each cell in the table contains the costs of
    routing from the origin to the destination for that cell.

    For more information, see: https://developer.tomtom.com/routing-api/documentation/tomtom-maps/matrix-routing-v2/matrix-routing-v2-service
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""Notifications API."""

from ..api import BaseApi


class NotificationsApi(BaseApi):
    """Notifications API.

    The Notifications API and its services provides the ability to read and clear the history of sent notifications. If you wish to stop receiving
    any notifications from the Notifications API services, please contact our Support and provide the address you wish to unsubscribe. See the
    Contact Groups documentation page for further information on the Contact Groups API. TomTom's Notifications service is intended to manage
    communication from Maps APIs to users. The following means of contact are supported:

    For more information, see: https://developer.tomtom.com/notifications-api/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""Routing API."""

from ..api import BaseApi
from ..models import LatLon, LatLonList
from .models import (
    CalculatedReachableRangeResponse,
    CalculatedRouteResponse,
    CalculateReachableRangePostData,
    CalculateReachableRouteParams,
    CalculateRouteParams,
    CalculateRoutePostData,
)


class RoutingApi(BaseApi):
    """Routing API.

    TomTom Routing is a suite of web services designed for developers to use our latest scalable routing engine.

    For more information, see: https://developer.tomtom.com/routing-api/documentation/tomtom-maps/routing-service
    """

    async def get_calculate_route(
        self,
        *,
        locations: LatLonList,
        params: CalculateRouteParams | None = None,
    ) -> CalculatedRouteResponse:
        """Get calculate route.

        For more information, see: https://developer.tomtom.com/routing-api/documentation/tomtom-maps/calculate-route
        """
        response = await self.get(
            endpoint=f"/routing/1/calculateRoute/{locations.to_colon_separated()}/json",
            params=params,
        )

        return await response.deserialize(CalculatedRouteResponse)

    async def post_calculate_route(
        self,
        *,
        locations: LatLonList,
        params: CalculateRouteParams | None = None,
        data: CalculateRoutePostData,
    ) -> CalculatedRouteResponse:
        """Post calculate route.

        For more information, see: https://developer.tomtom.com/routing-api/documentation/tomtom-maps/calculate-route

        Args:
            locations (LatLonList): A list of locations to include in the route.
            params (CalculateRouteParams | None, optional): Additional parameters for route calculation. Defaults to None.
            data (CalculateRoutePostData): Data specifying route details and constraints.

        Returns:
            CalculatedRouteResponse: The response containing the calculated route details.
        """
        response = await self.post(
            endpoint=f"/routing/1/calculateRoute/{locations.to_colon_separated()}/json",
            params=params,
            data=data,
        )

        return await response.deserialize(CalculatedRouteResponse)

    async def get_calculate_reachable_range(
        self,
        *,
        origin: LatLon,
        params: CalculateReachableRouteParams | None = None,
    ) -> CalculatedReachableRangeResponse:
        """Get calculate reachable range.

        For more information, see: https://developer.tomtom.com/routing-api/documentation/tomtom-maps/calculate-reachable-range

        Args:
            origin (LatLon): The origin point from which the reachable range is calculated.
            params (CalculateReachableRouteParams | None, optional): Additional parameters for the calculation. Defaults to None.

        Returns:
            CalculatedReachableRangeResponse: The response containing the details of the reachable range from the origin.
        """
        response = await self.get(
            endpoint=f"/routing/1/calculateReachableRange/{origin.to_comma_separated()}/json",
            params=params,
        )

        return await response.deserialize(CalculatedReachableRangeResponse)

    async def post_calculate_reachable_range(
        self,
        *,
        origin: LatLon,
        params: CalculateReachableRouteParams | None = None,
        data: CalculateReachableRangePostData,
    ) -> CalculatedReachableRangeResponse:
        """Post calculate reachable range.

        For more information, see: https://developer.tomtom.com/routing-api/documentation/tomtom-maps/calculate-reachable-range

        Args:
            origin (LatLon): The origin point from which the reachable range is calculated.
            params (CalculateReachableRouteParams | None, optional): Additional parameters for the calculation. Defaults to None.
            data (CalculateReachableRangePostData): Data specifying details for the reachable range calculation.

        Returns:
            CalculatedReachableRangeResponse: The response containing the details of the reachable range from the origin.
        """
        response = await self.post(
            endpoint=f"/routing/1/calculateReachableRange/{origin.to_comma_separated()}/json",
            params=params,
            data=data,
        )

        return await response.deserialize(CalculatedReachableRangeResponse)



"""Junction Analytics API."""

from ..api import BaseApi


class JunctionAnalyticsApi(BaseApi):
    """Junction Analytics API.

    Our Junction Analytics API provides input for efficient signal operations that makes it possible to allocate green time in a better way and
    reduce traffic delay. This service is designed for traffic signal hardware and software vendors who want to optimize signal operations and
    optimize traffic flows at intersections.

    For more information, see: https://developer.tomtom.com/junction-analytics/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""Models for the TomTom Traffic API."""
# pylint: disable=invalid-name, too-many-instance-attributes, too-many-lines

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum, StrEnum

from geojson import LineString, Point
from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from ..api import BaseParams, BasePostData
from ..models import Language, LatitudeLongitude, TileSizeType


@dataclass(kw_only=True)
class Aci:
    """Represents an ACI item."""

    probabilityOfOccurrence: ProbabilityOfOccurrenceType
    numberOfReports: int
    lastReportTime: str


@dataclass(kw_only=True)
class BBoxParam:
    """bbox param."""

    minLon: float
    minLat: float
    maxLon: float
    maxLat: float

    def to_comma_separated(self) -> str:
        """Convert the object into a comma-separated string."""
        return ",".join(map(str, [self.minLon, self.minLat, self.maxLon, self.maxLat]))


@dataclass(kw_only=True)
class BoudingBoxParam:
    """Boudingbox param."""

    minY: float
    minX: float
    maxY: float
    maxX: float

    def to_comma_separated(self) -> str:
        """Convert the object into a comma-separated string."""
        return ",".join(map(str, [self.minY, self.minX, self.maxY, self.maxX]))


class CategoryFilterType(StrEnum):
    """Supported category filter types."""

    UNKNOWN = "Unknown"
    ACCIDENT = "Accident"
    FOG = "Fog"
    DANGEROUS_CONDITIONS = "DangerousConditions"
    RAIN = "Rain"
    ICE = "Ice"
    JAM = "Jam"
    LANE_CLOSED = "LaneClosed"
    ROAD_CLOSED = "RoadClosed"
    ROAD_WORKS = "RoadWorks"
    WIND = "Wind"
    FLOODING = "Flooding"
    BROKEN_DOWN_VEHICLE = "BrokenDownVehicle"


@dataclass(kw_only=True)
class Coordinates:
    """Represents a coordinate item."""

    coordinate: list[LatitudeLongitude]


class DirectionType(StrEnum):
    """Supported direction types."""

    POSITIVE = "positive"
    NEGATIVE = "negative"


@dataclass(kw_only=True)
class Event:
    """Represents an event item."""

    description: str
    code: int
    iconCategory: IconCategoryType


@dataclass(kw_only=True)
class FlowSegmentData:
    """Represents a flow segment data item."""

    version: str = field(metadata=field_options(alias="@version"))
    frc: FrcType
    currentSpeed: int
    freeFlowSpeed: int
    currentTravelTime: int
    freeFlowTravelTime: float
    confidence: float
    coordinates: Coordinates
    openlr: str | None = None
    roadClosure: bool


@dataclass(kw_only=True)
class FlowSegmentDataParams(BaseParams):
    """Parameters for the get_flow_segment_data method."""

    unit: SpeedUnitType | None = None
    thickness: ThicknessType | None = None
    openLr: bool | None = None


@dataclass(kw_only=True)
class FlowSegmentDataResponse(DataClassORJSONMixin):
    """Represents a flow segment data response."""

    flowSegmentData: FlowSegmentData


class FlowStyleType(StrEnum):
    """Supported flow style types."""

    ABSOLUTE = "absolute"
    RELATIVE = "relative"
    RELATIVE0 = "relative0"
    RELATIVE0_DARK = "relative0-dark"
    RELATIVE_DELAY = "relative-delay"
    REDUCED_SENSITIVITY = "reduced-sensitivity"


class FlowTagType(StrEnum):
    """Supported flow tag types."""

    ROAD_TYPE = "road_type"
    TRAFFIC_LEVEL = "traffic_level"
    TRAFFIC_ROAD_COVERAGE = "traffic_road_coverage"
    LEFT_HAND_TRAFFIC = "left_hand_traffic"
    ROAD_CLOSURE = "road_closure"
    ROAD_CATEGORY = "road_category"
    ROAD_SUBCATEGORY = "road_subcategory"


class FlowType(StrEnum):
    """Supported flow types."""

    ABSOLUTE = "absolute"
    RELATIVE = "relative"
    RELATIVE_DELAY = "relative-delay"


class FrcType(StrEnum):
    """Supported Functional Road Class types. This indicates the road type."""

    FRC0 = "FRC0"  # Motorway, freeway or other major road
    FRC1 = "FRC1"  # Major road, less important than a motorway
    FRC2 = "FRC2"  # Other major road
    FRC3 = "FRC3"  # Secondary road
    FRC4 = "FRC4"  # Local connecting road
    FRC5 = "FRC5"  # Local road of high importance
    FRC6 = "FRC6"  # Local road


class IconCategoryType(IntEnum):
    """Supported icon category types."""

    UNKNOWN = 0
    ACCIDENT = 1
    FOG = 2
    DANGEROUS_CONDITIONS = 3
    RAIN = 4
    ICE = 5
    JAM = 6
    LANE_CLOSED = 7
    ROAD_CLOSED = 8
    ROAD_WORKS = 9
    WIND = 10
    FLOODING = 11
    BROKEN_DOWN_VEHICLE = 14


@dataclass(kw_only=True)
class Incident:
    """Represents an incident item."""

    type: str
    properties: IncidentProperties
    geometry: Point | LineString


@dataclass(kw_only=True)
class IncidentDetailsParams(BaseParams):
    """Parameters for the get_incident_details method."""

    fields: str | None = None
    language: Language | None = None
    t: str | None = None
    categoryFilter: list[CategoryFilterType] | None = None
    timeValidityFilter: list[TimeValidityFilterType] | None = None


@dataclass(kw_only=True)
class IncidentDetailsPostData(BasePostData):
    """Data for the post incident details API."""

    ids: list[str]


@dataclass(kw_only=True)
class IncidentDetailsResponse(DataClassORJSONMixin):
    """Represents an incident details response."""

    incidents: list[Incident | None]


@dataclass(kw_only=True)
class IncidentProperties:
    """Represents an incidents properties item."""

    id: str | None = None
    iconCategory: IconCategoryType
    magnitudeOfDelay: MagnitudeOfDelayType | None = None
    events: list[Event] | None = None
    startTime: str | None = None
    endTime: str | None = None
    from_: str | None = field(metadata=field_options(alias="from"), default=None)
    to: str | None = None
    length: float | None = None
    delay: int | None = None
    roadNumbers: list[str] | None = None
    timeValidity: str | None = None
    tmc: TMC | None = None
    probabilityOfOccurrence: ProbabilityOfOccurrenceType | None = None
    numberOfReports: int | None = None
    lastReportTime: str | None = None
    aci: Aci | None = None


class IncidentStyleType(StrEnum):
    """Supported incident style types."""

    S0 = "s0"
    S0_DARK = "s0-dark"
    S1 = "s1"
    S2 = "s2"
    S3 = "s3"
    NIGHT = "night"


class IncidentTagType(StrEnum):
    """Supported incident tag types."""

    ICON_CATEGORY = "icon_category"
    DESCRIPTION = "description"
    DELAY = "delay"
    ROAD_TYPE = "road_type"
    LEFT_HAND_TRAFFIC = "left_hand_traffic"
    MAGNITUDE = "magnitude"
    TRAFFIC_ROAD_COVERAGE = "traffic_road_coverage"
    CLUSTERED = "clustered"
    PROBABILITY_OF_OCCURRENCE = "probability_of_occurrence"
    NUMBER_OF_REPORTS = "number_of_reports"
    LAST_REPORT_TIME = "last_report_time"
    END_DATE = "end_date"
    ID = "id"
    ROAD_CATEGORY = "road_category"
    ROAD_SUBCATEGORY = "road_subcategory"


class IncidentTileFormatType(StrEnum):
    """Supported incident tile formats."""

    GIF = "gif"
    PNG = "png"


@dataclass(kw_only=True)
class IncidentViewportResponse(DataClassORJSONMixin):
    """Represents an incident viewport response."""

    viewpResp: ViewpResp


class MagnitudeOfDelayType(IntEnum):
    """Supported magnitude of delay types."""

    UNKNOWN = 0
    MINOR = 1
    MODERATE = 2
    MAJOR = 3
    UNDEFINED = 4


class ProbabilityOfOccurrenceType(StrEnum):
    """Supported probability of occurrence types."""

    CERTAIN = "certain"
    PROBABLE = "probable"
    RISK_OF = "risk_of"
    IMPROBABLE = "improbable"


@dataclass(kw_only=True)
class RasterFlowTilesParams(BaseParams):
    """Parameters for the get_raster_flow_tiles method."""

    thickness: ThicknessType | None = None
    tileSize: TileSizeType | None = None


@dataclass(kw_only=True)
class RasterIncidentTilesParams(BaseParams):
    """Parameters for the get_raster_incident_tile method."""

    t: str | None = None
    tileSize: TileSizeType | None = None


class RoadType(IntEnum):
    """Supported road types."""

    MOTORWAY = 0
    INTERNATIONAL_ROAD = 1
    MAJOR_ROAD = 2
    SECONDARY_ROAD = 3
    CONNECTING_ROAD = 4
    MAJOR_LOCAL_ROAD = 5
    LOCAL_ROAD = 6
    MINOR_LOCAL_ROAD = 7
    OTHER_ROADS = 8  # Non public road, Parking road, etc.


class SpeedUnitType(StrEnum):
    """Supported speed unit types."""

    KMPH = "kmph"
    MPH = "mph"


@dataclass(kw_only=True)
class TMC:
    """Represents a Traffic Message Channel item."""

    countryCode: str | None = None
    tableNumber: str | None = None
    tableVersion: str | None = None
    direction: DirectionType | None = None
    points: list[TmcPoint] | None = None


class ThicknessType(StrEnum):
    """Supported thickness types."""

    ABSOLUTE = "absolute"
    RELATIVE = "relative"
    RELATIVE_DELAY = "relative-delay"
    REDUCED_SENSITIVITY = "reduced-sensitivity"


class TimeValidityFilterType(StrEnum):
    """Supported time validity filter types."""

    PRESENT = "present"
    FUTURE = "future"


@dataclass(kw_only=True)
class TmcPoint:
    """Represents a TmcPoint item."""

    location: int | None = None
    offset: int


@dataclass(kw_only=True)
class TrafficState:
    """Represents a traffic state item."""

    trafficAge: str = field(metadata=field_options(alias="@trafficAge"))
    trafficModelId: str = field(metadata=field_options(alias="@trafficModelId"))


@dataclass(kw_only=True)
class VectorFlowTilesParams(BaseParams):
    """Parameters for the get_vector_flow_tiles method."""

    roadTypes: RoadType | None = None
    trafficLevelStep: float | None = None
    margin: float | None = None
    tags: list[FlowTagType] | None = None


@dataclass(kw_only=True)
class VectorIncidentTilesParams(BaseParams):
    """Parameters for the get_vector_incident_tile method."""

    t: str | None = None
    tags: list[str] | None = None
    language: Language | None = None


@dataclass(kw_only=True)
class ViewpResp:
    """Represents a view resp item."""

    trafficState: TrafficState
    copyrightIds: str
    version: str = field(metadata=field_options(alias="@version"))
    maps: str = field(metadata=field_options(alias="@maps"))



"""Models for the TomTom Routing API."""
# pylint: disable=invalid-name, too-many-instance-attributes, too-many-lines

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum

from mashumaro.mixins.orjson import DataClassORJSONMixin

from ..api import BaseParams, BasePostData
from ..models import AdrCategoryType, Language, LatitudeLongitude, TravelModeType


class AvoidType(StrEnum):
    """Supported avoid types."""

    TOLL_ROADS = "tollRoads"
    MOTORWAYS = "motorways"
    FERRIES = "ferries"
    UNPAVED_ROADS = "unpavedRoads"
    CARPOOLS = "carpools"
    ALREADY_USED_ROADS = "alreadyUsedRoads"
    BORDER_CROSSINGS = "borderCrossings"
    TUNNELS = "tunnels"
    CAR_TRAINS = "carTrains"
    LOW_EMISSION_ZONES = "lowEmissionZones"


@dataclass(kw_only=True)
class CalculateLongDistanceEVRouteParams(BaseParams):
    """Parameters for the calculate long distance EV route API."""

    vehicleHeading: int | None = None
    sectionType: SectionType | None = None
    report: str | None = None
    departAt: str | None = None
    routeType: RouteType | None = None
    traffic: bool | None = None
    avoid: AvoidType | None = None
    travelMode: TravelModeType | None = None
    vehicleMaxSpeed: int | None = None
    vehicleWeight: int | None = None
    vehicleAxleWeight: int | None = None
    vehicleNumberOfAxles: int | None = None
    vehicleLength: float | None = None
    vehicleWidth: float | None = None
    vehicleHeight: float | None = None
    vehicleCommercial: bool | None = None
    vehicleLoadType: VehicleLoadType | None = None
    vehicleAdrTunnelRestrictionCode: AdrCategoryType | None = None
    vehicleEngineType: VehicleEngineType
    accelerationEfficiency: float | None = None
    decelerationEfficiency: float | None = None
    uphillEfficiency: float | None = None
    downhillEfficiency: float | None = None
    consumptionInkWhPerkmAltitudeGain: float | None = None
    recuperationInkWhPerkmAltitudeLoss: float | None = None
    constantSpeedConsumptionInkWhPerHundredkm: str | None = None
    currentChargeInkWh: float | None = None
    maxChargeInkWh: float | None = None
    auxiliaryPowerInkW: float | None = None
    minChargeAtDestinationInkWh: float | None = None
    minChargeAtChargingStopsInkWh: float | None = None


@dataclass(kw_only=True)
class CalculateLongDistanceEVRoutePostData(BasePostData):
    """Data for the post calculate long distance EV route API."""

    chargingModes: list[ChargingMode]


@dataclass(kw_only=True)
class CalculateReachableRangePostData(BasePostData):
    """Data for the post calculate reachable range API."""

    avoidVignette: list[str] | None = None
    allowVignette: list[str] | None = None
    avoidAreas: Rectangles | None = None


@dataclass(kw_only=True)
class CalculateReachableRouteParams(BaseParams):
    """Parameters for the calculate reachable route API."""

    fuelBudgetInLiters: float | None = None
    energyBudgetInkWh: float | None = None
    timeBudgetInSec: float | None = None
    report: str | None = None
    departAt: str | None = None
    arriveAt: str | None = None
    routeType: RouteType | None = None
    traffic: bool | None = None
    avoid: AvoidType | None = None
    travelMode: TravelModeType | None = None
    hilliness: HillinessType | None = None
    windingness: WindingnessType | None = None
    vehicleMaxSpeed: int | None = None
    vehicleWeight: int | None = None
    vehicleAxleWeight: int | None = None
    vehicleNumberOfAxles: int | None = None
    vehicleLength: float | None = None
    vehicleWidth: float | None = None
    vehicleHeight: float | None = None
    vehicleCommercial: bool | None = None
    vehicleLoadType: VehicleLoadType | None = None
    vehicleAdrTunnelRestrictionCode: AdrCategoryType | None = None
    constantSpeedConsumptionInLitersPerHundredkm: str | None = None
    currentFuelInLiters: float | None = None
    auxiliaryPowerInLitersPerHour: float | None = None
    fuelEnergyDensityInMJoulesPerLiter: float | None = None
    accelerationEfficiency: float | None = None
    decelerationEfficiency: float | None = None
    uphillEfficiency: float | None = None
    downhillEfficiency: float | None = None
    consumptionInkWhPerkmAltitudeGain: float | None = None
    recuperationInkWhPerkmAltitudeLoss: float | None = None
    currentChargeInkWh: float | None = None
    maxChargeInkWh: float | None = None
    auxiliaryPowerInkW: float | None = None
    vehicleEngineType: VehicleEngineType | None = None
    constantSpeedConsumptionInkWhPerHundredkm: str | None = None


@dataclass(kw_only=True)
class CalculateRouteParams(BaseParams):
    """Parameters for the calculate route API."""

    maxAlternatives: int | None = None
    instructionsType: InstructionsType | None = None
    language: Language | None = None
    computeBestOrder: bool | None = None
    routeRepresentation: str | None = None
    computeTravelTimeFor: str | None = None
    vehicleHeading: int | None = None
    sectionType: SectionType | None = None
    report: str | None = None
    departAt: str | None = None
    arriveAt: str | None = None
    routeType: RouteType | None = None
    traffic: bool | None = None
    avoid: AvoidType | None = None
    travelMode: TravelModeType | None = None
    hilliness: HillinessType | None = None
    windingness: WindingnessType | None = None
    vehicleMaxSpeed: int | None = None  # 0-250
    vehicleWeight: int | None = None
    vehicleAxleWeight: int | None = None
    vehicleNumberOfAxles: int | None = None
    vehicleLength: float | None = None
    vehicleWidth: float | None = None
    vehicleHeight: float | None = None
    vehicleCommercial: bool | None = None
    vehicleLoadType: VehicleLoadType | None = None
    vehicleAdrTunnelRestrictionCode: AdrCategoryType | None = None
    vehicleEngineType: VehicleEngineType | None = None
    constantSpeedConsumptionInLitersPerHundredkm: str | None = None
    currentFuelInLiters: float | None = None
    auxiliaryPowerInLitersPerHour: float | None = None
    fuelEnergyDensityInMJoulesPerLiter: float | None = None
    accelerationEfficiency: float | None = None
    decelerationEfficiency: float | None = None
    uphillEfficiency: float | None = None
    downhillEfficiency: float | None = None
    consumptionInkWhPerkmAltitudeGain: float | None = None
    recuperationInkWhPerkmAltitudeLoss: float | None = None
    constantSpeedConsumptionInkWhPerHundredkm: str | None = None
    currentChargeInkWh: float | None = None
    maxChargeInkWh: float | None = None
    auxiliaryPowerInkW: float | None = None


@dataclass(kw_only=True)
class CalculateRoutePostData(BasePostData):
    """Data for the post calculate route API."""

    supportingPoints: list[LatitudeLongitude] | None = None
    avoidVignette: list[str] | None = None
    allowVignette: list[str] | None = None
    avoidAreas: Rectangles | None = None


@dataclass(kw_only=True)
class CalculatedLongDistanceEVRouteResponse(DataClassORJSONMixin):
    """Represents a calculated long distance EV route response."""

    formatVersion: str
    routes: list[EVRoute]


@dataclass(kw_only=True)
class CalculatedReachableRangeResponse(DataClassORJSONMixin):
    """Represents a calculated reachable range response."""

    formatVersion: str
    reachableRange: ReachableRange


@dataclass(kw_only=True)
class CalculatedRouteResponse(DataClassORJSONMixin):
    """Represents a calculated route response."""

    formatVersion: str
    routes: list[Route]


@dataclass(kw_only=True)
class ChargingConnection(DataClassORJSONMixin):
    """Represents a charging connection."""

    facilityType: FacilityType
    plugType: PlugType


@dataclass(kw_only=True)
class ChargingConnectionInfo(DataClassORJSONMixin):
    """Represents the charging connection info."""

    chargingVoltageInV: int
    chargingCurrentInA: int
    chargingCurrentType: ChargingCurrentType
    chargingPowerInkW: int
    chargingPlugType: PlugType


class ChargingCurrentType(StrEnum):
    """Supported charging current types."""

    DIRECT_CURRENT = "Direct_Current"
    ALTERNATING_CURRENT_1_PHASE = "Alternating_Current_1_Phase"
    ALTERNATING_CURRENT_3_PHASE = "Alternating_Current_3_Phase"


@dataclass(kw_only=True)
class ChargingCurve(DataClassORJSONMixin):
    """Represents a charging curve."""

    chargeInkWh: float
    timeToChargeInSeconds: int


@dataclass(kw_only=True)
class ChargingInformationAtEndOfLeg(DataClassORJSONMixin):
    """Represents the charging information at the end of a leg."""

    chargingConnections: list[ChargingConnection]
    chargingConnectionInfo: ChargingConnectionInfo
    targetChargeInkWh: float
    chargingTimeInSeconds: int
    chargingParkUuid: str
    chargingParkExternalId: str
    chargingParkName: str
    chargingParkOperatorName: str
    chargingParkLocation: ChargingParkLocation
    chargingParkPaymentOptions: list[ChargingParkPaymentOption]
    chargingParkPowerInkW: int
    chargingStopType: ChargingStopType


@dataclass(kw_only=True)
class ChargingMode(DataClassORJSONMixin):
    """Represents a charging mode."""

    chargingConnections: list[ChargingConnection]
    chargingCurve: list[ChargingCurve]


@dataclass(kw_only=True)
class ChargingParkLocation(DataClassORJSONMixin):
    """Represents a charging park location."""

    coordinate: LatitudeLongitude
    street: str
    city: str
    postalCode: str
    countryCode: str


@dataclass(kw_only=True)
class ChargingParkPaymentOption(DataClassORJSONMixin):
    """Represents a charging park payment option."""

    method: str
    brands: list[str]


class ChargingStopType(StrEnum):
    """Supported charging stop types."""

    AUTO_GENERATED = "Auto_Generated"
    USER_DEFINED = "User_Defined"


@dataclass(kw_only=True)
class EVLeg(DataClassORJSONMixin):
    """Represents a leg of a EV route."""

    summary: EVLegSummary | Summary
    points: list[LatitudeLongitude]


@dataclass(kw_only=True)
class EVLegSummary(DataClassORJSONMixin):
    """Represents the summary of a EV leg."""

    lengthInMeters: int
    travelTimeInSeconds: int
    trafficDelayInSeconds: int
    trafficLengthInMeters: int
    departureTime: datetime
    arrivalTime: datetime
    batteryConsumptionInkWh: float
    remainingChargeAtArrivalInkWh: float
    chargingInformationAtEndOfLeg: ChargingInformationAtEndOfLeg


@dataclass(kw_only=True)
class EVRoute(DataClassORJSONMixin):
    """Represents a EV route."""

    summary: EVSummary
    legs: list[EVLeg]
    sections: list[Section]


@dataclass(kw_only=True)
class EVSummary(DataClassORJSONMixin):
    """Represents the EV summary of a route."""

    lengthInMeters: int
    travelTimeInSeconds: int
    trafficDelayInSeconds: int
    trafficLengthInMeters: int
    departureTime: datetime
    arrivalTime: datetime
    batteryConsumptionInkWh: float
    remainingChargeAtArrivalInkWh: float
    totalChargingTimeInSeconds: float


class FacilityType(StrEnum):
    """Supported facility types."""

    BATTERY_EXCHANGE = "Battery_Exchange"
    CHARGE_100_TO_120V_1_PHASE_AT_8A = "Charge_100_to_120V_1_Phase_at_8A"
    CHARGE_100_TO_120V_1_PHASE_AT_10A = "Charge_100_to_120V_1_Phase_at_10A"
    CHARGE_100_TO_120V_1_PHASE_AT_12A = "Charge_100_to_120V_1_Phase_at_12A"
    CHARGE_100_TO_120V_1_PHASE_AT_13A = "Charge_100_to_120V_1_Phase_at_13A"
    CHARGE_100_TO_120V_1_PHASE_AT_16A = "Charge_100_to_120V_1_Phase_at_16A"
    CHARGE_100_TO_120V_1_PHASE_AT_32A = "Charge_100_to_120V_1_Phase_at_32A"
    CHARGE_200_TO_240V_1_PHASE_AT_8A = "Charge_200_to_240V_1_Phase_at_8A"
    CHARGE_200_TO_240V_1_PHASE_AT_10A = "Charge_200_to_240V_1_Phase_at_10A"
    CHARGE_200_TO_240V_1_PHASE_AT_12A = "Charge_200_to_240V_1_Phase_at_12A"
    CHARGE_200_TO_240V_1_PHASE_AT_16A = "Charge_200_to_240V_1_Phase_at_16A"
    CHARGE_200_TO_240V_1_PHASE_AT_20A = "Charge_200_to_240V_1_Phase_at_20A"
    CHARGE_200_TO_240V_1_PHASE_AT_32A = "Charge_200_to_240V_1_Phase_at_32A"
    CHARGE_200_TO_240V_1_PHASE_ABOVE_32A = "Charge_200_to_240V_1_Phase_above_32A"
    CHARGE_200_TO_240V_3_PHASE_AT_16A = "Charge_200_to_240V_3_Phase_at_16A"
    CHARGE_200_TO_240V_3_PHASE_AT_32A = "Charge_200_to_240V_3_Phase_at_32A"
    CHARGE_380_TO_480V_3_PHASE_AT_16A = "Charge_380_to_480V_3_Phase_at_16A"
    CHARGE_380_TO_480V_3_PHASE_AT_32A = "Charge_380_to_480V_3_Phase_at_32A"
    CHARGE_380_TO_480V_3_PHASE_AT_63A = "Charge_380_to_480V_3_Phase_at_63A"
    CHARGE_50_TO_500V_DIRECT_CURRENT_AT_62A_25KW = "Charge_50_to_500V_Direct_Current_at_62A_25kW"
    CHARGE_50_TO_500V_DIRECT_CURRENT_AT_125A_50KW = "Charge_50_to_500V_Direct_Current_at_125A_50kW"
    CHARGE_200_TO_450V_DIRECT_CURRENT_AT_200A_90KW = "Charge_200_to_450V_Direct_Current_at_200A_90kW"
    CHARGE_200_TO_480V_DIRECT_CURRENT_AT_255A_120KW = "Charge_200_to_480V_Direct_Current_at_255A_120kW"
    CHARGE_DIRECT_CURRENT_AT_20KW = "Charge_Direct_Current_at_20kW"
    CHARGE_DIRECT_CURRENT_AT_50KW = "Charge_Direct_Current_at_50kW"
    CHARGE_DIRECT_CURRENT_ABOVE_50KW = "Charge_Direct_Current_above_50kW"


class HillinessType(StrEnum):
    """Supported hilliness types."""

    LOW = "low"
    HIGH = "high"


class InstructionsType(StrEnum):
    """Supported instructions types."""

    CODED = "coded"
    TEXT = "text"
    TAGGED = "tagged"


@dataclass(kw_only=True)
class Leg(DataClassORJSONMixin):
    """Represents a leg of a route."""

    summary: Summary
    points: list[LatitudeLongitude]


class PlugType(StrEnum):
    """Supported plug types."""

    SMALL_PADDLE_INDUCTIVE = "Small_Paddle_Inductive"
    LARGE_PADDLE_INDUCTIVE = "Large_Paddle_Inductive"
    IEC_60309_1_PHASE = "IEC_60309_1_Phase"
    IEC_60309_3_PHASE = "IEC_60309_3_Phase"
    IEC_62196_TYPE_1_OUTLET = "IEC_62196_Type_1_Outlet"
    IEC_62196_TYPE_2_OUTLET = "IEC_62196_Type_2_Outlet"
    IEC_62196_TYPE_3_OUTLET = "IEC_62196_Type_3_Outlet"
    IEC_62196_TYPE_1_CONNECTOR_CABLE_ATTACHED = "IEC_62196_Type_1_Connector_Cable_Attached"
    IEC_62196_TYPE_2_CONNECTOR_CABLE_ATTACHED = "IEC_62196_Type_2_Connector_Cable_Attached"
    IEC_62196_TYPE_3_CONNECTOR_CABLE_ATTACHED = "IEC_62196_Type_3_Connector_Cable_Attached"
    COMBO_TO_IEC_62196_TYPE_1_BASE = "Combo_to_IEC_62196_Type_1_Base"
    COMBO_TO_IEC_62196_TYPE_2_BASE = "Combo_to_IEC_62196_Type_2_Base"
    TYPE_E_FRENCH_STANDARD_CEE_7_5 = "Type_E_French_Standard_CEE_7_5"
    TYPE_F_SCHUKO_CEE_7_4 = "Type_F_Schuko_CEE_7_4"
    TYPE_G_BRITISH_STANDARD_BS_1363 = "Type_G_British_Standard_BS_1363"
    TYPE_J_SWISS_STANDARD_SEV_1011 = "Type_J_Swiss_Standard_SEV_1011"
    CHINA_GB_PART_2 = "China_GB_Part_2"
    CHINA_GB_PART_3 = "China_GB_Part_3"
    IEC_309_DC_PLUG = "IEC_309_DC_Plug"
    AVCON_CONNECTOR = "AVCON_Connector"
    TESLA_CONNECTOR = "Tesla_Connector"
    NEMA_5_20 = "NEMA_5_20"
    CHADEMO = "CHAdeMO"
    SAE_J1772 = "SAE_J1772"
    TEPCO = "TEPCO"
    BETTER_PLACE_SOCKET = "Better_Place_Socket"
    MARECHAL_SOCKET = "Marechal_Socket"
    STANDARD_HOUSEHOLD_COUNTRY_SPECIFIC = "Standard_Household_Country_Specific"


@dataclass(kw_only=True)
class ReachableRange(DataClassORJSONMixin):
    """Represents a reachable range."""

    center: LatitudeLongitude
    boundary: list[LatitudeLongitude]


@dataclass(kw_only=True)
class Rectangle:
    """A rectangle defined by its south-west and north-east corners."""

    southWestCorner: LatitudeLongitude
    northEastCorner: LatitudeLongitude


@dataclass(kw_only=True)
class Rectangles:
    """A list of rectangles."""

    rectangles: list[Rectangle]


@dataclass(kw_only=True)
class Route(DataClassORJSONMixin):
    """Represents a route."""

    summary: Summary
    legs: list[Leg]
    sections: list[Section]


class RouteType(StrEnum):
    """Supported route types."""

    FASTEST = "fastest"
    SHORTEST = "shortest"
    SHORT = "short"
    ECO = "eco"
    THRILLING = "thrilling"


@dataclass(kw_only=True)
class Section(DataClassORJSONMixin):
    """Represents a section of a route."""

    startPointIndex: int
    endPointIndex: int
    sectionType: SectionReponseType
    travelMode: TravelModeType


class SectionReponseType(StrEnum):
    """Supported section reponse types."""

    CAR_TRAIN = "CAR_TRAIN"
    COUNTRY = "COUNTRY"
    FERRY = "FERRY"
    MOTORWAY = "MOTORWAY"
    PEDESTRIAN = "PEDESTRIAN"
    TOLL = "TOLL"
    TRAFFIC = "TRAFFIC"
    TRAVEL_MODE = "TRAVEL_MODE"
    TUNNEL = "TUNNEL"
    CARPOOL = "CARPOOL"
    URBAN = "URBAN"
    UNPAVED = "UNPAVED"
    LOW_EMISSION_ZONE = "LOW_EMISSION_ZONE"
    SPEED_LIMIT = "SPEED_LIMIT"


class SectionType(StrEnum):
    """Supported section types."""

    CAR_TRAIN = "carTrain"
    FERRY = "ferry"
    TUNNEL = "tunnel"
    MOTORWAY = "motorway"
    PEDESTRIAN = "pedestrian"
    TOLL_ROAD = "tollRoad"
    TOLL_VIGNETTE = "tollVignette"
    COUNTRY = "country"
    TRAVEL_MODE = "travelMode"
    TRAFFIC = "traffic"


@dataclass(kw_only=True)
class Summary(DataClassORJSONMixin):
    """Represents the summary of a route."""

    lengthInMeters: int
    travelTimeInSeconds: int
    trafficDelayInSeconds: int
    trafficLengthInMeters: int
    departureTime: datetime
    arrivalTime: datetime


class VehicleEngineType(StrEnum):
    """Supported vehicle engine types."""

    COMBUSTION = "combustion"
    ELECTRIC = "electric"


class VehicleLoadType(StrEnum):
    """Supported vehicle load types."""

    US_HAZMAT_CLASS_1 = "USHazmatClass1"  # Explosives (USA)
    US_HAZMAT_CLASS_2 = "USHazmatClass2"  # Compressed gas (USA)
    US_HAZMAT_CLASS_3 = "USHazmatClass3"  # Flammable liquids (USA)
    US_HAZMAT_CLASS_4 = "USHazmatClass4"  # Flammable solids (USA)
    US_HAZMAT_CLASS_5 = "USHazmatClass5"  # Oxidizers (USA)
    US_HAZMAT_CLASS_6 = "USHazmatClass6"  # Poisons (USA)
    US_HAZMAT_CLASS_7 = "USHazmatClass7"  # Radioactive (USA)
    US_HAZMAT_CLASS_8 = "USHazmatClass8"  # Corrosives (USA)
    US_HAZMAT_CLASS_9 = "USHazmatClass9"  # Miscellaneous (USA)
    OTHER_HAZMAT_EXPLOSIVE = "otherHazmatExplosive"  # Explosives (Rest of the world)
    OTHER_HAZMAT_GENERAL = "otherHazmatGeneral"  # Miscellaneous (Rest of the world)
    OTHER_HAZMAT_HARMFUL_TO_WATER = "otherHazmatHarmfulToWater"  # Harmful to water (Rest of the world)


@dataclass(kw_only=True)
class WaypointOptimizationOptions:
    """Options for the waypoint optimization API."""

    travelMode: TravelModeType
    vehicleMaxSpeed: int
    vehicleWeight: int
    vehicleAxleWeight: int
    vehicleLength: float
    vehicleWidth: float
    vehicleHeight: float
    vehicleCommercial: bool
    vehicleLoadType: list[VehicleLoadType]
    vehicleAdrTunnelRestrictionCode: AdrCategoryType


@dataclass(kw_only=True)
class WaypointOptimizationPoint:
    """A waypoint optimization point."""

    point: LatitudeLongitude


@dataclass(kw_only=True)
class WaypointOptimizationPostData(BasePostData):
    """Data for the post waypoint optimization API."""

    waypoints: list[WaypointOptimizationPoint]
    options: WaypointOptimizationOptions


@dataclass(kw_only=True)
class WaypointOptimizedResponse(DataClassORJSONMixin):
    """Represents a waypoint optimized response."""

    optimizedOrder: list[int]


class WindingnessType(StrEnum):
    """Supported windingness types."""

    LOW = "low"
    HIGH = "high"



"""Intermediate Traffic API."""

from ..api import BaseApi


class IntermediateTrafficApi(BaseApi):
    """Intermediate Traffic API.

    The mission of TomTom is helping our customers arrive at their destinations faster, more safely and more reliably, regardless of their locations.
    We developed TomTom Intermediate Traffic to deliver detailed, real-time traffic content to business customers who integrate it into their own
    applications. Target customers for TomTom Intermediate Traffic include automotive OEMs, web and application developers, and governments. We
    deliver bulk traffic flow information that provides a comprehensive view of the entire road network.

    For more information, see: https://developer.tomtom.com/intermediate-traffic-service/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""Models for the TomTom Maps API."""
# pylint: disable=invalid-name, too-many-instance-attributes, too-many-lines

from __future__ import annotations

from dataclasses import dataclass, field
from enum import StrEnum

from mashumaro.mixins.orjson import DataClassORJSONMixin

from ..api import BaseParams
from ..models import AdrCategoryType, Language, TileSizeType, TravelModeType, ViewType
from ..utils import serialize_list_brackets


class DangerousGoodsLoadType(StrEnum):
    """Supported dangerous good types."""

    EXPLOSIVES = "Explosives"
    GASES = "Gases"
    FLAMMABLE_LIQUIDS = "Flammable_Liquids"
    FLAMMABLE_SOLIDS = "Flammable_Solids"
    OXIDIZING_AND_ORGANIC_SUBSTANCE = "Oxidizing_And_Organic_Substance"
    TOXIC_AND_INFECTIOUS_SUBSTANCE = "Toxic_And_Infectious_Substance"
    RADIOACTIVE_MATERIAL = "Radioactive_Material"
    CORROSIVES = "Corrosives"
    MISCELLANEOUS_DANGEROUS_GOODS = "Miscellaneous_Dangerous_Goods"


class EmissionClassType(StrEnum):
    """Supported emission class types."""

    EMISSIONCLASS0 = "EmissionClass0"
    EMISSIONCLASS1 = "EmissionClass1"
    EMISSIONCLASS2 = "EmissionClass2"
    EMISSIONCLASS3 = "EmissionClass3"
    EMISSIONCLASS4 = "EmissionClass4"
    EMISSIONCLASS5 = "EmissionClass5"
    EMISSIONCLASS6 = "EmissionClass6"
    EMISSIONCLASS7 = "EmissionClass7"
    EMISSIONCLASS8 = "EmissionClass8"
    EMISSIONCLASS9 = "EmissionClass9"
    EMISSIONCLASS10 = "EmissionClass10"


class EngineType(StrEnum):
    """Supported engine types."""

    LPG = "LPG"
    CNG = "CNG"
    LNG = "LNG"
    DIESEL = "Diesel"
    PETROL = "Petrol"
    HYDROGEN = "Hydrogen"
    ELECTRIC = "Electric"
    HYBRID = "Hybrid"
    PLUGIN_HYBRID = "Plugin_Hybrid"


class GeneralLoadType(StrEnum):
    """Supported general load types."""

    GENERAL_HAZARDOUS_MATERIALS = "General_Hazardous_Materials"
    EXPLOSIVE_MATERIALS = "Explosive_Materials"
    GOODS_HARMFUL_TO_WATER = "Goods_Harmful_To_Water"


class IncludeType(StrEnum):
    """Supported include types."""

    ROAD_RESTRICTIONS = "road_restrictions"


class LayerType(StrEnum):
    """Supported layer types."""

    BASIC = "basic"
    HYBRID = "hybrid"
    LABELS = "labels"


class LayerTypeWithPoiType(StrEnum):
    """Supported layer types."""

    BASIC = "basic"
    HYBRID = "hybrid"
    LABELS = "labels"
    POI = "poi"


@dataclass(kw_only=True)
class MapServiceCopyrightsResponse(DataClassORJSONMixin):
    """Represents the map service copyrights response."""

    formatVersion: str
    copyrightsCaption: str


@dataclass(kw_only=True)
class MapTileParams(BaseParams):
    """Parameters for the map tile API."""

    tileSize: TileSizeType | None = None
    view: ViewType | None = None
    language: Language | None = None


@dataclass(kw_only=True)
class MapTileV1Params(BaseParams):
    """Parameters for the map tile API."""

    view: ViewType | None = None
    language: Language | None = None


@dataclass(kw_only=True)
class MapTileV2Params(BaseParams):
    """Parameters for the map tile API."""

    view: ViewType | None = None
    include: list[IncludeType] | None = field(default=None, metadata={"serialize": serialize_list_brackets})
    vehicleWeight: int | None = None
    vehicleAxleWeight: int | None = None
    numberOfAxles: int | None = None
    vehicleLength: float | None = None
    vehicleWidth: float | None = None
    vehicleHeight: float | None = None
    generalLoadType: list[GeneralLoadType] | None = field(default=None, metadata={"serialize": serialize_list_brackets})
    dangerousGoodsLoadType: list[DangerousGoodsLoadType] | None = field(default=None, metadata={"serialize": serialize_list_brackets})
    adrCategory: AdrCategoryType | None = None
    commercialVehicle: bool | None = None
    travelMode: TravelModeType | None = None
    emissionClass: list[EmissionClassType] | None = field(default=None, metadata={"serialize": serialize_list_brackets})
    engineType: list[EngineType] | None = field(default=None, metadata={"serialize": serialize_list_brackets})
    travelModeProfile: str | None = None


@dataclass(kw_only=True)
class StaticImageParams(BaseParams):
    """Parameters for the map tile API."""

    layer: LayerType | None = None
    style: StyleType | None = None
    x: int | None = None
    y: int | None = None
    zoom: int | None = None
    center: list[float] | None = None
    format: TileFormatType | None = None
    width: int | None = None  # must be a positive integer between 1 and 8192.
    height: int | None = None  # must be a positive integer between 1 and 8192.
    bbox: list[float] | None = None
    view: ViewType | None = None


class StyleType(StrEnum):
    """Supported style types."""

    MAIN = "main"
    NIGHT = "night"


class TileFormatType(StrEnum):
    """Supported tile formats."""

    PNG = "png"
    JPG = "jpg"



"""Route Monitoring API."""

from ..api import BaseApi


class RouteMonitoringApi(BaseApi):
    """Rote Monitoring API.

    The TomTom Route Monitoring API service provides an intuitive and powerful way to monitor strategic routes in real-time. Customers have the
    ability to pre-define routes important to their businesses getting detailed information on current travel time, current delay time and percentage
    delay, route distance, live data coverage, and data confidence level. The overall route information can also be checked on a segment level,
    providing accurate and detailed information of traffic flow dynamics down to short length extensions.

    For more information, see: https://developer.tomtom.com/route-monitoring/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError

    # There are no methods defined on the developer portal.



"""Snap to Roads API."""

from ..api import BaseApi


class SnapToRoadsApi(BaseApi):
    """Snap to Roads API.

    The TomTom Snap to Roads API and service offers a solution to enable you to get the most out of your applications, and grants you the ability to
    use advanced map data. It is a web service designed for developers to create web and mobile applications responsible for matching received points
    (gathered from GPS devices) to map a road network and reconstruct the road driven by a customer, and provides detailed information about the
    matched route. These web services can be used via REST APIs.

    For more information, see: https://developer.tomtom.com/snap-to-roads-api/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""Traffic Stats API."""

from ..api import BaseApi


class TrafficStatsApi(BaseApi):
    """Traffic Stats API.

    Traffic Stats is a suite of web services designed for developers to create web applications which analyse historical traffic data. These web
    services are RESTful APIs. The Traffic Stats APIs are based on the collection of Floating Car Data (FCD), which is a proven and innovative method
    of measuring what is happening on the road.

    For more information, see: https://developer.tomtom.com/traffic-stats/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""Geofencing API."""

from ..api import BaseApi


class GeofencingApi(BaseApi):
    """Geofencing API.

    TomTom's Geofencing API service is intended to define virtual barriers on real geographical locations. Together with the location of an object,
    you can determine whether that object is located within, outside, or close to a predefined geographical area.

    For more information, see: https://developer.tomtom.com/geofencing-api/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""Reverse Geocode API."""

from ..api import BaseApi
from ..models import LatLon
from ..places.models import CrossStreetLookupParams, ReverseGeocodeParams, ReverseGeocodeResponse


class ReverseGeocodingApi(BaseApi):
    """Reverse Geocoding API.

    The TomTom Reverse Geocoding API gives users a tool to translate a coordinate (for example: 37.786505, -122.3862) into a human-understandable
    street address, street element, or geography. Most often, this is needed in tracking applications where you receive a GPS feed from the device or
    asset and you want to know the address.

    For more information, see: https://developer.tomtom.com/reverse-geocoding-api/documentation/product-information/introduction
    """

    async def get_reverse_geocode(
        self,
        *,
        position: LatLon,
        params: ReverseGeocodeParams | None = None,
    ) -> ReverseGeocodeResponse:
        """Get reverse geocode.

        For more information, see: https://developer.tomtom.com/reverse-geocoding-api/documentation/reverse-geocode

        Args:
            position (LatLon): The latitude and longitude of the location to reverse geocode.
            params (ReverseGeocodeParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            ReverseGeocodeResponse: The response containing the reverse geocode results.
        """
        response = await self.get(
            endpoint=f"/search/2/reverseGeocode/{position.to_comma_separated()}.json",
            params=params,
        )

        return await response.deserialize(ReverseGeocodeResponse)

    async def get_cross_street_lookup(
        self,
        *,
        position: LatLon,
        params: CrossStreetLookupParams | None = None,
    ) -> ReverseGeocodeResponse:
        """Get cross street lookup.

        For more information, see: https://developer.tomtom.com/reverse-geocoding-api/documentation/cross-street-lookup

        Args:
            position (LatLon): The latitude and longitude of the location for the cross street lookup.
            params (CrossStreetLookupParams | None, optional): Additional parameters for the request. Defaults to None.

        Returns:
            ReverseGeocodeResponse: The response containing the cross street lookup results.
        """
        response = await self.get(
            endpoint=f"/search/2/reverseGeocode/crossStreet/{position.to_comma_separated()}.json",
            params=params,
        )

        return await response.deserialize(ReverseGeocodeResponse)



"""Parking Availability API."""

from ..api import BaseApi
from ..automotive.models import ParkingAvailabilityParams, ParkingAvailabilityResponse


class ParkingAvailabilityApi(BaseApi):
    """Parking Availability API.

    For more information, see: https://developer.tomtom.com/parking-availability-api/documentation/product-information/introduction
    """

    async def get_parking_availability(
        self,
        *,
        params: ParkingAvailabilityParams | None = None,
    ) -> ParkingAvailabilityResponse:
        """Get parking availability.

        For more information, see: https://developer.tomtom.com/parking-availability-api/documentation/parking-availability-api/parking-availability

        Args:
            params (ParkingAvailabilityParams | None, optional): Additional parameters for the parking availability. Defaults to None.

        Returns:
            ParkingAvailabilityResponse: Response containing parking availability data.
        """
        response = await self.get(
            endpoint="/search/2/parkingAvailability.json",
            params=params,
        )

        return await response.deserialize(ParkingAvailabilityResponse)



"""Location History API."""

from ..api import BaseApi


class LocationHistoryApi(BaseApi):
    """Location History API.

    TomTom's Location History API is intended to keep track and manage the locations of multiple objects. It can share data with TomTom's Geofencing
    service to enhance it with the history of object transitions through fence borders.

    For more information, see: https://developer.tomtom.com/location-history-api/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""O/D Analysis API."""

from ..api import BaseApi


class ODAnalysisApi(BaseApi):
    """O/D Analysis API.

    O/D Analysis is a suite of web services designed for developers to create web applications which analyze historical traffic data. These web
    services are RESTful APIs. The O/D Analysis API services are based on the collection of Floating Car Data (FCD), which is a proven and innovative
    method of measuring what is happening on the road.

    For more information, see: https://developer.tomtom.com/od-analysis/documentation/product-information/introduction
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """Not implemented.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError



"""Routing APIs."""

from .long_distance_ev_routing import LongDistanceEVRoutingApi
from .matrix_routing_v2 import MatrixRoutingApiV2
from .routing import RoutingApi
from .waypoint_optimization import WaypointOptimizationApi

__all__ = [
    "LongDistanceEVRoutingApi",
    "MatrixRoutingApiV2",
    "RoutingApi",
    "WaypointOptimizationApi",
]



"""Generic utils."""

import math
from enum import Enum, IntEnum, StrEnum

from .models import LatLon, MapTile

# Constants
MIN_ZOOM_LEVEL = 0
MAX_ZOOM_LEVEL = 22
MIN_LAT = -85.051128779807
MAX_LAT = 85.051128779806
MIN_LON = -180.0
MAX_LON = 180.0


def lat_lon_to_tile_zxy(lat: float, lon: float, zoom_level: int) -> MapTile:
    """Convert a location to a map tile for a given zoom level.

    For more information, see: https://developer.tomtom.com/map-display-api/documentation/zoom-levels-and-tile-grid

    Args:
        lat (float): The latitude of the location.
        lon (float): The longitude of the location.
        zoom_level (int): The zoom level.

    Returns:
        MapTile: The corresponding map tile.

    Raises:
        ValueError: If latitude, longitude, or zoom level are out of range.

    """
    if not MIN_ZOOM_LEVEL <= zoom_level <= MAX_ZOOM_LEVEL:
        raise ValueError(f"Zoom level value is out of range [{MIN_ZOOM_LEVEL}, {MAX_ZOOM_LEVEL}]")

    if not MIN_LAT <= lat <= MAX_LAT:
        raise ValueError(f"Latitude value is out of range [{MIN_LAT}, {MAX_LAT}]")

    if not MIN_LON <= lon <= MAX_LON:
        raise ValueError(f"Longitude value is out of range [{MIN_LON}, {MAX_LON}]")

    z = zoom_level
    xy_tiles_count = 2**z
    x = int(((lon + 180.0) / 360.0) * xy_tiles_count)
    y = int(((1.0 - math.log(math.tan(math.radians(lat)) + 1.0 / math.cos(math.radians(lat))) / math.pi) / 2.0) * xy_tiles_count)

    return MapTile(x=x, y=y, zoom=z)


def tile_zxy_to_lat_lon(zoom_level: int, x: int, y: int) -> LatLon:
    """Convert a map tile to a location for a given zoom level.

    For more information, see: https://developer.tomtom.com/map-display-api/documentation/zoom-levels-and-tile-grid

    Args:
        zoom_level (int): The zoom level.
        x (int): The x coordinate of the map tile.
        y (int): The y coordinate of the map tile.

    Returns:
        LatLon: The corresponding latitude and longitude.

    Raises:
        ValueError: If the zoom level or tile coordinates are out of range.

    """
    if not MIN_ZOOM_LEVEL <= zoom_level <= MAX_ZOOM_LEVEL:
        raise ValueError(f"Zoom level value is out of range [{MIN_ZOOM_LEVEL}, {MAX_ZOOM_LEVEL}]")

    z = zoom_level
    max_xy = 2**z - 1

    if not 0 <= x <= max_xy:
        raise ValueError(f"Tile x value is out of range [0, {max_xy}]")

    if not 0 <= y <= max_xy:
        raise ValueError(f"Tile y value is out of range [0, {max_xy}]")

    lon = (x / 2**z) * 360.0 - 180.0

    n = math.pi - (2.0 * math.pi * y) / 2**z
    lat = (180.0 / math.pi) * math.atan(0.5 * (math.exp(n) - math.exp(-n)))

    return LatLon(lat=lat, lon=lon)


def serialize_bool(x: bool) -> str:
    """Serialize a boolean as a lowercase string.

    Args:
        x (bool): The boolean to serialize.

    Returns:
        str: The serialized boolean as a lowercase string.
    """
    return str(x).lower()


def serialize_enum(x: Enum) -> str:
    """Serialize an Enum as a string.

    If the Enum is an IntEnum, serialize it as a string.
    If the Enum is a StrEnum, serialize it as itself.
    Otherwise, serialize it as its value.
    """
    if isinstance(x, IntEnum):
        return str(x)
    if isinstance(x, StrEnum):
        return x
    return str(x.value)


def serialize_list(x: list[int | float | bool | str | Enum]) -> str | None:
    """Serialize a list to a comma-separated string, converting booleans to lowercase strings.

    Args:
        x (list[int | float | bool | str | Enum]): The list to serialize.

    Returns:
        str | None: The serialized list as a comma-separated string, or None if the list was empty.
    """
    if not x:
        return None
    return ",".join(serialize_bool(item) if isinstance(item, bool) else serialize_enum(item) if isinstance(item, Enum) else str(item) for item in x)


def serialize_list_brackets(x: list[int | float | bool | str | Enum]) -> str | None:
    """Serialize a list to a comma-separated string, converting booleans to lowercase strings, and surround with square brackets.

    Args:
        x (list[int | float | bool | str | Enum]): The list to serialize.

    Returns:
        str | None: The serialized list as a comma-separated string, or None if the list was empty.
    """
    serialized_list = serialize_list(x)
    return f"[{serialized_list}]" if serialized_list is not None else None
